from inspect import getmodule
from collections import abc
from pathlib import Path
from itertools import product
from typing import Sequence, Optional, Union, overload, Tuple

from coba.evaluators import SequentialCB

from coba.pipes import Pipes, DiskSink, ListSink, DiskSource, ListSource, Identity, Insert
from coba.results import Result, TransactionDecode, TransactionEncode, TransactionResult
from coba.context import CobaContext, ExceptLog, StampLog, NameLog, DecoratedLogger, ExceptionLogger
from coba.exceptions import CobaException
from coba.multiprocessing import CobaMultiprocessor
from coba.primitives import Learner, Environment, Evaluator
from coba.utilities import PackageChecker

from coba.experiments.process import MakeTasks, ChunkTasks, ProcessTasks

class Experiment:
    """Experiment for environments, learners and evaluators."""

    @overload
    def __init__(self,
        environments : Union[Environment, Sequence[Environment]],
        learners     : Union[Learner,Sequence[Learner]],
        evaluator    : Union[Evaluator,Sequence[Evaluator]] = SequentialCB(),
        description  : str = None) -> None:
        """Instantiate an Experiment.

        Args:
            environments: The collection of environments to use in the experiment.
            learners: The collection of learners to use in the experiment.
            evaluator: The evaluation task we wish to perform on learners and environments.
            description: A description of the experiment for documentaiton purposes.
        """

    @overload
    def __init__(self,
        eval_tuples: Sequence[Union[Tuple[Learner,Environment],Tuple[Learner,Environment,Evaluator]]],
        description: str = None) -> None:
        ...
        """Instantiate an Experiment.

        Args:
            eval_tuples: The learner-environment-evaluator triples we wish to evaluate. If no evaluator
                is provided the default SequentialCB() will be used for evaluation.
            description: A description of the experiment for documentaiton purposes.
        """

    def __init__(self, *args,**kwargs) -> None:
        """Instantiate an Experiment.

        Args:
            environments: The collection of environments to use in the experiment.
            learners: The collection of learners to use in the experiment.
            evaluator: The evaluation task we wish to perform on learners and environments.
            eval_tuples: The learner-environment-evaluator triples we wish to evaluate.
            description: A description of the experiment for documentaiton purposes.
        """

        if 'evaluation_task' in kwargs:
            raise CobaException(
                "The `evaluation_task` argument has been deprecated. Instead of `Experiment(x,y,evaluation_task=t)` "
                "you should call `Experiment(x,y,t)`. Additionally if you are using a custom evaluator then the "
                "main evaluation method should be `def evaluate(self, learner, environment)` rather than "
                "`def process(self,learner,interactions)`."
            )

        args = list(args)
        if len(args) > 0 and not isinstance(args[0],(abc.Sequence,abc.Iterable)): args[0] = [args[0]]

        triples,description = self._parse_init_args(*args,**kwargs)

        self._triples     = triples
        self._description = description

        # if any([lrn is None for _,lrn,_ in self._triples]):
        #     raise CobaException("A Learner was given whose value was None, which can't be processed.")

        # if any([env is None for env,_,_ in self._triples]):
        #     raise CobaException("An Environment was given whose value was None, which can't be processed.")

        self._processes        : Optional[int] = None
        self._maxchunksperchild: Optional[int] = None
        self._maxtasksperchunk : Optional[int] = None

    def config(self,
        processes: int = None,
        maxchunksperchild: int = None,
        maxtasksperchunk: int = None) -> 'Experiment':
        """Configure how the experiment will be executed.

        Args:
            processes: The number of processes to create for evaluating the experiment.
            maxchunksperchild: The number of chunks each process evaluate before being restarted. A
                value of 0 means that all processes will survive until the end of the experiment.
            maxtasksperchunk: The maximum number of tasks a chunk can have. If a chunk has too many
                tasks it will be split into smaller chunks. A value of 0 means that chunks are never
                broken down into smaller chunks.

        Returns:
            The configured Experiment.
        """

        assert processes is None or processes > 0, "The given number of processes is invalid. Must be greater than 0."
        assert maxchunksperchild is None or maxchunksperchild >= 0, "The given number of chunks per child is invalid. Must be greater than or equal to 0 (0 for infinite)."
        assert maxtasksperchunk is None or maxtasksperchunk >= 0, "The given number of tasks per chunk is invalid. Must be greater than or equal to 0 (0 for infinite)."

        self._processes         = processes
        self._maxchunksperchild = maxchunksperchild
        self._maxtasksperchunk  = maxtasksperchunk

        return self

    @property
    def processes(self) -> int:
        """The number of processes to use when evaluating the experiment."""
        return self._processes if self._processes is not None else CobaContext.experiment.processes

    @property
    def maxchunksperchild(self) -> int:
        """The number of tasks chunks to perform per process before restarting an evaluation process."""
        return self._maxchunksperchild if self._maxchunksperchild is not None else CobaContext.experiment.maxchunksperchild

    @property
    def maxtasksperchunk(self) -> int:
        """The maximum number of tasks allowed in a chunk before breaking a chunk into smaller chunks."""
        return self._maxtasksperchunk if self._maxtasksperchunk is not None else CobaContext.experiment.maxtasksperchunk

    def run(self,
            result_file:str = None,
            quiet:bool = False,
            processes:int = None,
            maxchunksperchild: int = None,
            maxtasksperchunk: int = None,
            seed: Optional[int] = 1) -> Result:
        """Run the experiment and return the results.

        Args:
            result_file: The file for writing and restoring results.
            quiet: Indicates that logged output should be turned off.
            processes: The number of processes to create for evaluating the experiment.
            maxchunksperchild: The number of chunks each process evaluates before being restarted. A
                value of 0 means that processes will survive until the end of the experiment. Unless
                otherwise specified chunks are individual environment-learner-evaluator triples.
            maxtasksperchunk: The maximum number of tasks a chunk can have. If a chunk has too many
                tasks it will be split into smaller chunks. A value of 0 means that chunks are never
                broken down into smaller chunks.
            seed: The seed that will determine all randomness within the experiment.

        Returns:
            Result of the experiment.
        """

        self.config(processes,maxchunksperchild,maxtasksperchunk)
        mp,mc,mt = self.processes,self.maxchunksperchild,self.maxtasksperchunk

        CobaContext.store['experiment_seed'] = seed
        is_multiproc = mp > 1 or mc != 0

        if is_multiproc: self._check_for_cloudpickle_dependency()

        old_logger = CobaContext.logger

        if quiet:
            CobaContext.logger = DecoratedLogger([], ExceptionLogger(CobaContext.logger.sink), [NameLog(),StampLog()] if is_multiproc else [StampLog()])
        else:
            CobaContext.logger = DecoratedLogger([ExceptLog()], CobaContext.logger, [NameLog(),StampLog()] if is_multiproc else [StampLog()])

        CobaContext.logger.log("Experiment Started")

        if result_file and Path(result_file).exists():
            CobaContext.logger.log("Restoring Results")
            restored = Result.from_file(result_file)
        else:
            restored = None

        n_given_lrns = len(set([l for _,l,_ in self._triples]))
        n_given_envs = len(set([e for e,_,_ in self._triples]))

        meta = {'n_learners':n_given_lrns,'n_environments':n_given_envs,'description':self._description,'seed':seed}

        workitems = MakeTasks(self._triples,restored)
        chunker   = ChunkTasks(mt)
        process   = CobaMultiprocessor(ProcessTasks(), mp, mc, False)
        encode    = TransactionEncode(restored)
        sink      = DiskSink(result_file,batch=1) if result_file else ListSink(foreach=True)
        source    = DiskSource(result_file) if result_file else ListSource(sink.items)
        decode    = TransactionDecode()
        result    = TransactionResult()
        preamble  = Identity() if restored else Insert([["T0",meta]])

        try:
            lrn_mismatch = restored and n_given_lrns != restored.experiment.get('n_learners',n_given_lrns)
            env_mismatch = restored and n_given_envs != restored.experiment.get('n_environments',n_given_envs)
            if lrn_mismatch or env_mismatch: raise CobaException("The experiment does not match the given logs")
            Pipes.join(workitems, chunker, process, preamble, encode, sink).run()
        except KeyboardInterrupt: #pragma: no cover
            CobaContext.logger.log("Experiment Aborted (aborted via Ctrl-C)")
        except Exception as ex: #pragma: no cover
            CobaContext.logger.log(ex)
            CobaContext.logger.log("Experiment Failed")
        else:
            CobaContext.logger.log("Experiment Finished")

        CobaContext.logger = old_logger
        del CobaContext.store['experiment_seed']

        return Pipes.join(source,decode,result).read()

    def _parse_init_args(self,*args,**kwargs) -> Tuple[Sequence[Tuple[Environment,Learner]], Evaluator, Optional[str]]:
        #we know this with 100% certainty
        definite_triples = ({'eval_tuples'} & kwargs.keys()            ) or \
                           (len(args) == 1 and 'learners' not in kwargs) or \
                           (len(args) == 2 and isinstance(args[1],str) )
        try:
            if definite_triples:
                triples = args[0] if len(args) > 0 else kwargs['eval_tuples']
                descr   = args[1] if len(args) > 1 else kwargs.get('description',None)
                triples = [t if len(t) ==3 else (*t,SequentialCB()) for t in triples]
                return triples, descr

            else:
                envs  = args[0] if len(args) > 0 else kwargs['environments']
                lrns  = args[1] if len(args) > 1 else kwargs['learners']
                vals  = args[2] if len(args) > 2 else kwargs.get('evaluator',SequentialCB())
                descr = args[3] if len(args) > 3 else kwargs.get('description',None)
                if not isinstance(envs,abc.Sequence): envs = [envs]
                if not isinstance(lrns,abc.Sequence): lrns = [lrns]
                if not isinstance(vals,abc.Sequence): vals = [vals]
                return list(product(envs,lrns,vals)), descr

        except KeyError as e:
            raise TypeError(f'Experiment.__init__ missing required arguments')

    def _check_for_cloudpickle_dependency(self) -> None:
        objs = set()

        for e,l,v in self._triples:
            objs.add(l)
            objs.add(v)

            try:
                objs.update(e)
            except:
                objs.add(e)

        if any(not hasattr(getmodule(o),'__file__') for o in objs):
            PackageChecker.cloudpickle('Coba multiprocessing in Jupyter')



import multiprocessing as mp
from ctypes import c_short
from typing import Iterable, Any, Dict

from coba.utilities  import coba_exit, peek_first
from coba.context    import CobaContext, ConcurrentCacher, Logger, Cacher
from coba.primitives import Filter, Sink
from coba.pipes      import Multiprocessor, Foreach, QueueSink, QueueSource, ThreadLine, SourceSink

class CobaMultiprocessor(Filter[Iterable[Any], Iterable[Any]]):

    class ProcessFilter:

        def __init__(self, filter: Filter, logger: Logger, cacher: Cacher, store: Dict[str,Any], logger_sink: Sink) -> None:

            self._filter      = filter
            self._logger      = logger
            self._cacher      = cacher
            self._store       = store
            self._logger_sink = logger_sink

        def filter(self, item: Any) -> Any:

            #this is set inside the process
            CobaContext.logger = self._logger
            CobaContext.cacher = self._cacher
            CobaContext.store  = self._store

            #at this point logger has been marshalled so we can
            #modify it without affecting the base process logger
            CobaContext.logger.sink = self._logger_sink

            yield from self._filter.filter(item)

    def __init__(self, filter: Filter, processes:int=1, maxtasksperchild:int=0, chunked:bool=False) -> None:
        self._filter           = filter
        self._processes        = processes
        self._maxtasksperchild = maxtasksperchild
        self._chunked          = chunked

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:

        _, items = peek_first(items)
        if not items: return []

        try:
            if self._maxtasksperchild == 0 and self._processes == 1:
                filter = self._filter
            else:
                #There are three potential contexts -- spawn, fork, and forkserver. On windows and mac spawn is the only option.
                #On Linux the default option is fork. Fork creates processes faster and can share memory but also doesn't play
                #well with threads. Therefore, to make behavior consistent across Linux, Windows, and Mac and avoid potential bugs
                #we force our multiprocessors to always use spawn.
                spawn_context = mp.get_context("spawn")

                stdlog        = spawn_context.Queue()
                array         = spawn_context.RawArray(c_short,[0]*2**16)
                lock          = spawn_context.Lock()
                read_stdlog   = QueueSource(stdlog)
                write_stdlog  = QueueSink(stdlog)

                stdlog_writer = ThreadLine(SourceSink(read_stdlog,Foreach(CobaContext.logger.sink)))
                stdlog_writer.start()

                logger = CobaContext.logger
                cacher = ConcurrentCacher(CobaContext.cacher,array,lock)
                store  = { "openml_semaphore": spawn_context.Semaphore(3), **CobaContext.store }

                filter = CobaMultiprocessor.ProcessFilter(self._filter, logger, cacher, store, write_stdlog)

            try:
                yield from Multiprocessor(filter, self._processes, self._maxtasksperchild).filter(items)

            except Exception as e:
                # If the error was due to an uncaught exception in the given filter it could be the case that the user
                # is expecting it therefore we don't want to supress it. On the other hand, if the error is due to the
                # act of multiprocessing then we know the user is not expecting it and we log it in a friendly way.

                #I have since changed my mind. We should raise it regardless.
                raise

            finally:
                 if self._maxtasksperchild != 0 or self._processes > 1:
                    write_stdlog.write(None) #attempt to shutdown the logging process gracefully by sending the poison pill
                    stdlog_writer.join()
                    stdlog.close()

        except RuntimeError as e: #pragma: no cover
            #This happens when importing main causes this code to run again
            coba_exit(str(e))



import time
import traceback

from abc import abstractmethod, ABC
from multiprocessing import current_process
from contextlib import contextmanager, nullcontext
from datetime import datetime
from typing import ContextManager, Iterator, Sequence, Union
from copy import copy

from coba.primitives import Filter, Sink
from coba.exceptions import CobaException
from coba.pipes import Pipes, NullSink, ConsoleSink, Identity

class Logger(ABC):
    """The interface for a logger."""

    @property
    @abstractmethod
    def sink(self) -> Sink[str]:
        """The sink the logger writes to."""
        ...

    @sink.setter
    @abstractmethod
    def sink(self, sink: Sink[str]):
        ...

    @abstractmethod
    def log(self, message: Union[str,Exception]) -> 'ContextManager[Logger]':
        """Log a message or exception to the sink.

        Args:
            message: The message or exception that should be logged.

        Returns:
            A ContextManager that controls hierarchy.
        """
        ...

    @abstractmethod
    def time(self, message: str) -> 'ContextManager[Logger]':
        """Log a timed message to the sink.

        Args:
            message: The message that should be logged.

        Returns:
            A ContextManager that controls when timing stops.
        """
        ...

class NullLogger(Logger):
    """A logger which writes nothing."""

    def __init__(self, sink: Sink[str] = NullSink()) -> None:
        """Instantiate a NullLogger.

        Args:
            sink: The sink to write to (by default null).
        """
        self._sink = sink

    @property
    def sink(self) -> Sink[str]:
        return self._sink

    @sink.setter
    def sink(self, sink: Sink[str]):
        self._sink = sink

    def log(self, message: Union[str,Exception]) -> 'ContextManager[Logger]':
        return nullcontext(self)

    def time(self, message: str) -> 'ContextManager[Logger]':
        return nullcontext(self)

class BasicLogger(Logger):
    """A Logger with flat hierarchy."""

    def __init__(self, sink: Sink[str] = ConsoleSink()):
        """Instantiate a BasicLogger.

        Args:
            sink: The sink to write to (by default console).
        """
        self._sink = sink
        self._starts = []

    @contextmanager
    def _log_context(self, message:str) -> 'Iterator[Logger]':
        outcome = "(error)"
        try:
            yield self
            outcome = "(completed)"
        except KeyboardInterrupt:
            outcome = "(interrupt)"
            raise
        except (Exception, TypeError) as e:
            outcome = "(exception)"
            raise
        finally:
            self.log(f"{message} {outcome}")

    @contextmanager
    def _time_context(self, message:str) -> 'Iterator[Logger]':
        self.log(message)
        self._starts.append(time.time())
        try:
            yield self
        except KeyboardInterrupt:
            self.log(f"{message} ({round(time.time()-self._starts.pop(),2)} seconds) (interrupt)")
            raise
        except Exception:
            self.log(f"{message} ({round(time.time()-self._starts.pop(),2)} seconds) (exception)")
            raise
        else:
            self.log(f"{message} ({round(time.time()-self._starts.pop(),2)} seconds) (completed)")

    @property
    def sink(self) -> Sink[str]:
        return self._sink

    @sink.setter
    def sink(self, sink: Sink[str]):
        self._sink = sink

    def log(self, message: str) -> 'ContextManager[Logger]':
        self._sink.write(message)
        return self._log_context(message)

    def time(self, message: str) -> 'ContextManager[Logger]':
        return self._time_context(message)

class IndentLogger(Logger):
    """A Logger with indented hierarchy."""

    def __init__(self, sink: Sink[str] = ConsoleSink()):
        """Instantiate an IndentLogger.

        Args:
            sink: The sink to write the logs to (by default console).
        """
        self._sink = sink

        self._messages = []
        self._level    = 0
        self._bullets  = dict(enumerate(['','* ','> ','- ','+ ']))

    @contextmanager
    def _indent_context(self) -> 'Iterator[Logger]':
        try:
            self._level += 1
            yield self
        finally:
            self._level -= 1

    @contextmanager
    def _time_context(self, message:str) -> 'Iterator[Logger]':

        # we don't have all the information we need to write
        # our message but we want to save our place in line
        # we also level our message before entering context
        place_in_line = len(self._messages)
        self._messages.append("Placeholder")
        message = self._level_message(message)

        with self._indent_context():
            outcome = "(error)"
            try:
                start = time.time()
                yield self
                outcome = "(completed)"
            except KeyboardInterrupt:
                outcome = "(interrupt)"
                raise
            except (Exception, TypeError):
                outcome = "(exception)"
                raise
            finally:

                self._messages[place_in_line] = message + f" ({round(time.time()-start,2)} seconds) {outcome}"

                if place_in_line == 0:
                    while self._messages:
                        self._sink.write(self._messages.pop(0))

    def _level_message(self, message: str) -> str:
        indent = '  ' * self._level
        bullet = self._bullets.get(self._level,'~')

        return indent + bullet + message

    @property
    def sink(self) -> Sink[str]:
        return self._sink

    @sink.setter
    def sink(self, sink: Sink[str]):
        self._sink = sink

    def log(self, message: Union[str,Exception]) -> 'ContextManager[Logger]':
        if self._messages:
            self._messages.append(self._level_message(message))
        else:
            self._sink.write(self._level_message(message))

        return self._indent_context()

    def time(self, message: str) -> 'ContextManager[Logger]':
        return self._time_context(message)

class ExceptionLogger(Logger):
    """A Logger that only logs exceptions."""

    def __init__(self, sink: Sink[str] = ConsoleSink()):
        """Instantiate an ExceptionLogger.

        Args:
            sink: The sink to write the logs to (by default console).
        """
        self._sink = sink
        self._filter = ExceptLog()

    @property
    def sink(self) -> Sink[str]:
        return self._sink

    @sink.setter
    def sink(self, sink: Sink[str]):
        self._sink = sink

    def log(self, message: Union[str,Exception]) -> 'ContextManager[Logger]':
        if isinstance(message,Exception):
            self._sink.write(self._filter.filter(message))
        return nullcontext(self)

    def time(self, message: str) -> 'ContextManager[Logger]':
        return nullcontext(self)

class DecoratedLogger(Logger):
    """A Logger which decorates a logger."""

    def __init__(self, pre_decorators: Sequence[Filter], logger: Logger, post_decorators: Sequence[Filter]):
        """Instantiate DecoratedLogger.

        Args:
            pre_decorators: A sequence of decorators to be applied before the base logger.
            logger: The base logger we are decorating.
            post_decorators: A sequence of decorators to be applied after the base logger.
        """

        self._pre_decorator    = Pipes.join(*pre_decorators) if pre_decorators else Identity()
        self._post_decorators  = post_decorators
        self._original_logger  = logger
        self._copy_logger      = copy(logger)
        self._copy_logger.sink = Pipes.join(*post_decorators, logger.sink)

    @property
    def sink(self) -> Sink[str]:
        return self._original_logger.sink

    @sink.setter
    def sink(self, sink: Sink[str]):
        self._original_logger.sink = sink
        self._copy_logger.sink     = Pipes.join(*self._post_decorators, sink)

    def log(self, message: Union[str,Exception]) -> 'ContextManager[Logger]':
        return self._copy_logger.log(self._pre_decorator.filter(message))

    def time(self, message: str) -> 'ContextManager[Logger]':
        return self._copy_logger.time(self._pre_decorator.filter(message))

    def undecorate(self) -> Logger:
        """Remove the decorator.

        Returns:
            The original logger without a decorator.
        """
        return self._original_logger

class NameLog(Filter[str,str]):
    """Add process name to logs."""

    def filter(self, log: str) -> str:
        return f"pid-{current_process().pid:<6} -- {log}"

class StampLog(Filter[str,str]):
    """Add timestamp to logs."""

    def filter(self, log: str) -> str:
        return f"{self._now().strftime('%Y-%m-%d %H:%M:%S')} -- {log}"

    def _now(self)-> datetime:
        return datetime.now()

class ExceptLog(Filter[Union[str,Exception],str]):
    """Add stack traces to logs."""

    def filter(self, log: Union[str,Exception]) -> str:
        if isinstance(log, str):
            return log
        elif isinstance(log, CobaException):
            return f"EXCEPTION: {log}"
        else:
            return f'Unexpected exception:\n\n{self.format_exception(log)}'

    def format_exception(self, ex: Exception):

        out = []

        if ex.__cause__ is not None:
            out.append(self.format_exception(ex.__cause__))
            out.append("The above exception was the direct cause of the following exception:\n")

        msg = ''.join(traceback.TracebackException.from_exception(ex).format_exception_only())

        #we manualy format to provide more helpful messages for strange edge conditions
        #the alternative method is tb  = ''.join(traceback.format_tb(ex.__traceback__))
        tbs  = []
        for frame in traceback.extract_tb(ex.__traceback__):

            code   = frame.line or '<unknown code, this is likely due to the code being in a Jupyter cell>'
            fname  = frame.filename or '<unknown file>'
            mname  = frame.name or '<unknown method>'
            lineno = frame.lineno or '<unknown line>'

            tbs.append(f'  File "{fname}", line {lineno}, in {mname}\n    {code}\n')
        tb = ''.join(tbs)

        if tb : out.append(tb)
        if msg: out.append(f'  {msg}')

        return '\n'.join(out)



import warnings
import math

from collections import defaultdict
from typing import Any, Mapping, Optional, Hashable, Tuple, Sequence

from coba.random import CobaRandom
from coba.primitives import Context, Action, Actions, Prob
from coba.statistics import OnlineVariance
from coba.primitives import Dense, Sparse, HashableDense, HashableSparse, Learner
from coba.learners.utilities import PMFPredictor

def make_hashable(item):
    if isinstance(item,Dense): return HashableDense(item)
    if isinstance(item,Sparse): return HashableSparse(item)
    return item

class BanditEpsilonLearner(Learner):
    """Select the greedy action with probability (1-epsilon)."""

    def __init__(self, epsilon: float=.05, seed:int = 1) -> None:
        """Instantiate an BanditEpsilonLearner.

        Args:
            epsilon: We explore with probability epsilon and exploit otherwise.
            seed: The seed used to select actions in predict.
        """
        self._epsilon = epsilon
        self._N: Mapping[Hashable, int            ] = defaultdict(int)
        self._Q: Mapping[Hashable, Optional[float]] = defaultdict(int)
        self._pred = PMFPredictor(self._pmf,seed)

    @property
    def params(self) -> Mapping[str, Any]:
        return {'family': 'BanditEpsilon', 'epsilon': self._epsilon, 'seed': self._pred.seed}

    def _pmf(self, context, actions):

        actions     = list(map(make_hashable,actions))
        values      = [ self._Q[action] for action in actions ]
        max_value   = None if set(values) == {None} else max(v for v in values if v is not None)
        max_indexes = [i for i in range(len(values)) if values[i]==max_value]

        prob_selected_randomly = [1/len(actions) * self._epsilon] * len(actions)
        prob_selected_greedily = [ int(i in max_indexes)/len(max_indexes) * (1-self._epsilon) for i in range(len(actions))]

        return [p1+p2 for p1,p2 in zip(prob_selected_randomly,prob_selected_greedily)]

    def score(self, context: Context, actions: Actions, action: Action) -> Prob:
        return self._pred.score(context,actions,action)

    def predict(self, context: Context, actions: Actions) -> Tuple['Action','Prob']:
        return self._pred.predict(context,actions)

    def learn(self, context: 'Context', action: 'Action', reward: float, probability: float) -> None:
        action = make_hashable(action)
        alpha = 1/(self._N[action]+1)
        old_Q = self._Q[action] or 0.0
        self._Q[action] = (1-alpha) * old_Q + alpha * reward
        self._N[action] = self._N[action] + 1

class BanditUCBLearner(Learner):
    """Select the action with the highest upper confidence bound estimate.

    This algorithm is an implementation of Auer et al. (2002) UCB1-Tuned algorithm.
    The paper's proven regret bounds only assume that rewards have support in [0,1].

    References:
        Auer, Peter, Nicolo Cesa-Bianchi, and Paul Fischer. "Finite-time analysis of
        the multiarmed bandit problem." Machine learning 47.2-3 (2002): 235-256.
    """

    def __init__(self, seed: int = 1):
        """Instantiate a BanditUcbLearner.

        Args:
            seed: The seed used to select actions in predict.
        """
        #these variable names were selected for easier comparison with the original paper
        self._t: int = 0
        self._m: Mapping['Action', float         ] = {}
        self._s: Mapping['Action', int           ] = {}
        self._v: Mapping['Action', OnlineVariance] = {}
        self._pred = PMFPredictor(self._pmf,seed)

    @property
    def params(self) -> Mapping[str, Any]:
        return {'family': 'BanditUCB', 'seed': self._pred.seed }

    def _pmf(self, context, actions):

        actions = list(map(make_hashable,actions))
        never_observed_actions = set(actions) - self._m.keys()

        if never_observed_actions:
            max_actions = never_observed_actions
        else:
            values      = [ self._m[a] + self._Avg_R_UCB(a) for a in actions ]
            max_value   = max(values)
            max_actions = [ a for a,v in zip(actions,values) if v==max_value ]

        return [int(action in max_actions)/len(max_actions) for action in actions]

    def score(self, context: Context, actions: Actions, action: Action) -> Prob:
        return self._pred.score(context,actions,action)

    def predict(self, context: Context, actions: Actions) -> Tuple['Action','Prob']:
        return self._pred.predict(context,actions)

    def learn(self, context: 'Context', action: 'Action', reward: float, probability: float) -> None:

        self._t += 1

        action = make_hashable(action)

        if action not in self._m:
            self._m[action] = reward
            self._s[action] = 1
            self._v[action] = OnlineVariance()

        else:
            self._m[action] = (1-1/self._s[action]) * self._m[action] + 1/self._s[action] * reward
            self._s[action] += 1
            self._v[action].update(reward)

    def _Avg_R_UCB(self, action: 'Action') -> float:
        """Produce the estimated upper confidence bound (UCB) for E[R|A].

        Args:
            action: The action for which we want to retrieve UCB for E[R|A].

        Returns:
            The estimated UCB for E[R|A].

        Remarks:
            See the beginning of section 4 in the algorithm's paper for this equation.
        """
        ln = math.log; n = self._t; n_j = self._s[action]; V_j = self._Var_R_UCB(action)

        return math.sqrt(ln(n)/n_j * min(1/4,V_j))

    def _Var_R_UCB(self, action: 'Action') -> float:
        """Produce the upper confidence bound (UCB) for Var[R|A].

        Args:
            action: The action for which we want to retrieve UCB for Var[R|A].

        Returns:
            The estimated UCB for Var[R|A].

        Remarks:
            See the beginning of section 4 in the algorithm's paper for this equation.
        """
        ln = math.log; t = self._t; s = self._s[action]; var = self._v[action].variance

        return var + math.sqrt(2*ln(t)/s)

class FixedLearner(Learner):
    """Select actions from a fixed distribution and learn nothing."""

    def __init__(self, pmf: Sequence['Prob'], seed: int = 1) -> None:
        """Instantiate a FixedLearner.

        Args:
            pmf: A PMF whose values are the probability of taking each action.
            seed: The seed used to select actions in predict.
        """
        assert round(sum(pmf),3) == 1, "The given pmf must sum to one to be a valid pmf."
        assert all([p >= 0 for p in pmf]), "All given probabilities of the pmf must be greater than or equal to 0."
        self._fpmf = pmf
        self._pred = PMFPredictor(self._pmf,seed)

    @property
    def params(self) -> Mapping[str, Any]:
        return {'family': 'fixed', 'seed': self._pred.seed}

    def _pmf(self, context, actions):
        return self._fpmf

    def score(self, context: Context, actions: Actions, action: Action) -> Prob:
        return self._pred.score(context,actions,action)

    def predict(self, context: Context, actions: Actions) -> Tuple['Action','Prob']:
        return self._pred.predict(context,actions)

    def learn(self, context: 'Context', action: 'Action', reward: float, prob: float) -> None:
        pass

class RandomLearner(Learner):
    """Select actions from a uniform distribution and learn nothing."""

    def __init__(self, seed:int = 1):
        """Instantiate a RandomLearner.

        Args:
            seed: The seed used to select actions in predict.
        """
        self._rng = CobaRandom(seed)

    @property
    def params(self) -> Mapping[str, Any]:
        return {'family':'Random', 'seed': self._rng.seed}

    def score(self, context: 'Context', actions: 'Actions', action: 'Action') -> 'Prob':
        return 1/len(actions)

    def predict(self, context: 'Context', actions: 'Actions') -> Tuple['Action','Prob']:
        return self._rng.choicew(actions)

    def learn(self, context: 'Context', action: 'Action', reward: float, probability: float) -> None:
        pass



import sys

if sys.version_info >= (3,10):#pragma: no cover
    from importlib.metadata import entry_points

else:#pragma: no cover
    def entry_points(*,group):
        import importlib.metadata
        return importlib.metadata.entry_points()[group]



from typing import Iterable, Mapping, Any

from coba.exceptions import CobaException
from coba.primitives import Source, Environment, Interaction

class ResultEnvironment(Environment):

    def __init__(self, int_source:Source[str], env_params:dict, lrn_params:dict, val_params:dict) -> None:
        self._int_source = int_source
        self._env_params = env_params
        self._lrn_params = lrn_params
        self._val_params = val_params

    @property
    def params(self) -> Mapping[str,Any]:
        params = {}
        if self._env_params: params.update(self._env_params)
        if self._lrn_params: params.update(self._lrn_params)
        if self._val_params: params.update(self._val_params)
        return params

    def read(self) -> Iterable[Interaction]:
        interactions = self._int_source.read()

        is_all_data = {'actions','rewards'}.issubset(interactions.keys())
        is_log_data = {'action' ,'reward' }.issubset(interactions.keys())

        if not (is_all_data or is_log_data):
            raise CobaException(
                "It is not possible to create a ResultEnvironment if the Result does "
                "not contain at least (`actions`,`rewards`) or (`action`,`reward`).")

        for i in range(len(next(iter(interactions.values())))):
            yield {k:interactions[k][i] for k in interactions.keys()}



from collections import Counter
from typing import Sequence

from coba.exceptions import CobaException
from coba.primitives import Pipe

def resolve_params(pipes: Sequence[Pipe]):

    params = []
    for p in pipes:

        if not hasattr(p,'params'):
            continue
        if callable(p.params):
            raise CobaException(f"The params on {type(p).__name__} is not decorated with @property")
        if not isinstance(p.params,dict):
            raise CobaException(f"The params on {type(p).__name__} is not a dict type")

        params.append(p.params)

    counts = Counter([k for p in params for k in p.keys()])
    index  = {}

    def resolve_key_conflicts(key):
        if counts[key] == 1:
            return key
        else:
            index[key] = index.get(key,0)+1
            return f"{key}{index[key]}"

    return { resolve_key_conflicts(k):v for p in params for k,v in p.items() }



import os
import gzip
import time

from hashlib import blake2b
from threading import Lock, current_thread
from pathlib import Path
from contextlib import nullcontext, contextmanager
from collections.abc import Iterator
from collections import defaultdict
from abc import abstractmethod, ABC
from typing import Union, Dict, TypeVar, Iterable, Optional, Callable, Generic, Sequence, ContextManager

from coba.exceptions import CobaException

_K = TypeVar("_K")
_V = TypeVar("_V")

class Cacher(Generic[_K, _V], ABC):
    """The interface for a cacher."""

    @abstractmethod
    def __contains__(self, key: _K) -> bool:
        """Determine if key is in cache."""
        ...

    @abstractmethod
    def rmv(self, key: _K) -> None:
        """Remove a key from the cache.

        Args:
            key: The key to remove from the cache.
        """
        ...

    @abstractmethod
    def get_set(self, key: _K, getter: Union[Callable[[], _V],_V]) -> ContextManager[_V]:
        """Get a key from the cache.

        If the key is not in the cache put it first using getter.

        Args:
            key: The key to get from the cache.
            getter: A method for getting the value if necessary.
        """
        ...

class NullCacher(Cacher[_K, _V]):
    """A cacher which does not cache any data."""

    def __init__(self) -> None:
        """Instantiate a NullCacher."""
        pass

    def __contains__(self, key: _K) -> bool:
        return False

    def rmv(self, key: _K):
        pass

    def get_set(self, key: _K, getter: Union[Callable[[], _V],_V]) -> ContextManager[_V]:
        return nullcontext(getter())

class MemoryCacher(Cacher[_K, _V]):
    """A cacher that caches in memory."""

    def __init__(self) -> None:
        """Instantiate a MemoryCacher."""
        self._cache: Dict[_K,_V] = {}

    def __contains__(self, key: _K) -> bool:
        return key in self._cache

    def rmv(self, key: _K) -> None:
        if key in self:
            del self._cache[key]

    def get_set(self, key: _K, getter: Union[Callable[[], _V],_V]) -> ContextManager[_V]:
        if key not in self:
            value = getter() if callable(getter) else getter
            value = list(value) if isinstance(value,Iterator) else value
            self._cache[key] = value
            return nullcontext(value)
        return nullcontext(self._cache[key])

class DiskCacher(Cacher[str, Iterable[str]]):
    """A cacher that writes to disk.

    The DiskCacher compresses all values before writing to conserve disk space.
    """

    def __init__(self, cache_dir: Union[str, Path]) -> None:
        """Instantiate a DiskCacher.

        Args:
            cache_dir: The directory path where all given keys will be cached as files
        """
        self.cache_directory = cache_dir

    @property
    def cache_directory(self) -> Optional[str]:
        """The directory where the cache will write to disk."""
        return str(self._cache_dir)

    @cache_directory.setter
    def cache_directory(self, cache_dir:Union[Path,str]) -> None:
        if not isinstance(cache_dir,(str,Path)):
            raise CobaException(f"An invalid cache directory was supplied: {cache_dir}.")
        self._cache_dir = str(cache_dir)

    def __contains__(self, key: str) -> bool:
        return self._cache_path(key).exists()

    def rmv(self, key: str) -> None:
        if self._cache_path(key).exists(): self._cache_path(key).unlink()

    def get_set(self, key: str, getter: Union[Callable[[], Iterable[str]],Iterable[str]]) -> ContextManager[Iterable[str]]:

        if key in self and os.path.getsize(self._cache_path(key)) == 0:
            self.rmv(key)

        if key not in self:
            try:
                lines = getter() if callable(getter) else getter
                if isinstance(lines,str): lines = [lines]
                Path(self._cache_dir).expanduser().mkdir(parents=True, exist_ok=True)

                with gzip.open(self._cache_path(key), "wt+", 6, "utf-8") as f:
                    for line in lines:
                        f.write(line.rstrip('\r\n'))
                        f.write('\n')
            except:
                if key in self:
                    self.rmv(key)
                raise

        return gzip.open(self._cache_path(key),mode='rt',encoding='utf-8')

    def _cache_name(self, key: str) -> str:
        if not all(c.isalnum() or c in (' ','.','_') for c in key):
            raise CobaException(f"A key was given to DiskCacher which couldn't be made into a file, {key}")
        return f"{key}.gz"

    def _cache_path(self, key: str) -> Path:
        return Path(self._cache_dir).expanduser()/self._cache_name(key)

class ConcurrentCacher(Cacher[_K, _V]):
    """A cacher that is multiprocess safe."""

    def __init__(self, cache:Cacher[_K, _V], list: Sequence = None, lock: Lock = None):
        """Instantiate a ConcurrentCacher.

        Args:
            cache: Base cacher to make read/write safe across multiple processes.
            list: Shared memory object that allows tracking of read/write locks.
            lock: Synchronization object to be used to ensure read/write safety.
        """
        self._digest_size = 2

        self._cache = cache
        self._lock  = lock or Lock()
        self._array = list or [0]*2**(8*self._digest_size)

        self._write_waits = 0 # for testing purposes only. won't be accurate in production.
        self._read_waits  = 0 # for testing purposes only. won't be accurate in production.

        self._locks = defaultdict(int) # for safety to make sure our current process/thread isn't waiting on itself

        assert len(self._array) >= 2**(8*self._digest_size)

    def __contains__(self, key: _K) -> bool:
        return key in self._cache

    def rmv(self, key: _K):
        lock = None
        try:
            if key in self:
                self._acquire_write_lock(key)
                lock='write'
                self._cache.rmv(key)
                lock = None
                self._release_write_lock(key)
        except:
            if lock == 'write': self._release_write_lock(key)
            raise

    def get_set(self, key: _K, getter: Union[Callable[[],_V],_V]) -> ContextManager[_V]:

        try:
            self._acquire_read_lock(key)
            if key in self._cache:
                return self._release_read_on_exit(key,self._cache.get_set(key,None))
            self._release_read_lock(key)

            self._acquire_write_lock(key)
            if key in self:#pragma: no cover; this is super hard to isolate so I'm just going to trust it...
                self._switch_write_to_read_lock(key)
                return self._release_read_on_exit(key,self._cache.get_set(key,None))
            else:
                item = self._cache.get_set(key, getter)
                self._switch_write_to_read_lock(key)
                return self._release_read_on_exit(key,item)
        except Exception as e:
            if self._has_read_lock(key): self._release_read_lock(key)
            if self._has_write_lock(key): self._release_write_lock(key)
            raise

    @contextmanager
    def _release_read_on_exit(self,key:str,item:ContextManager[_V]) -> _V:
        try:
            with item as out:
                yield out
        finally:
            self._release_read_lock(key)

    def _acquire_read_lock(self, key):
        if self._has_write_lock(key):
            raise CobaException("The concurrent cacher was asked to enter an unrecoverable state.")
        index = self._index(key)
        self._read_waits += 1
        while True:
            with self._lock:
                if self._array[index] >= 0:
                    self._locks[(current_thread().ident,key)] += 1
                    self._array[index] += 1
                    self._read_waits -= 1
                    break
            time.sleep(1)

    def _release_read_lock(self, key):
        index = self._index(key)
        with self._lock:
            self._array[index] -= 1
            self._locks[(current_thread().ident,key)] -= 1

    def _acquire_write_lock(self, key) -> ContextManager:
        if self._has_write_lock(key) or self._has_read_lock(key):
            raise CobaException("The concurrent cacher was asked to enter an unrecoverable state.")
        index = self._index(key)
        self._write_waits += 1
        while True:
            with self._lock:
                if self._array[index] == 0:
                    self._locks[(current_thread().ident,key)] = -1
                    self._array[index] = -1
                    self._write_waits -= 1
                    break
            time.sleep(1)

    def _release_write_lock(self, key):
        index = self._index(key)
        with self._lock:
            self._array[index] = 0
            self._locks[(current_thread().ident,key)] = 0

    def _switch_write_to_read_lock(self, key) -> None:
        index = self._index(key)
        assert self._array[index] == -1, "You don't have write permissions"
        assert self._locks[(current_thread().ident,key)] == -1, "You don't have write permissions"
        with self._lock:
            self._array[index] = 1
            self._locks[(current_thread().ident,key)] = 1

    def _has_read_lock(self, key) -> bool:
        return self._locks[(current_thread().ident,key)] > 0

    def _has_write_lock(self, key) -> bool:
        return self._locks[(current_thread().ident,key)] == -1

    def _index(self, key) -> int:
        return int.from_bytes(blake2b(str(key).encode('utf-8'),digest_size=self._digest_size).digest(),"big")



import time
import pickle
import warnings
import copy

from math import isnan
from statistics import median, stdev, mode
from numbers import Number
from zlib import crc32
from operator import eq, methodcaller, itemgetter
from collections import defaultdict
from functools import lru_cache
from itertools import islice, chain, tee, compress, repeat
from typing import Optional, Sequence, Union, Iterable, Any, Tuple, Callable, Mapping, Literal

from coba              import pipes, primitives
from coba.context      import CobaContext
from coba.random       import CobaRandom
from coba.exceptions   import CobaException
from coba.statistics   import iqr
from coba.utilities    import peek_first, PackageChecker, try_else, minimize
from coba.primitives   import is_batch, Learner, Interaction, EnvironmentFilter, BinaryReward, DiscreteReward
from coba.pipes        import Pipes, SparseDense
from coba.safety       import SafeLearner

class Identity(pipes.Identity, EnvironmentFilter):
    """Return interactions unchanged."""
    pass

class Take(pipes.Take, EnvironmentFilter):
    """Take the first n interactions."""
    pass

class Slice(pipes.Slice, EnvironmentFilter):
    """Take a slice of interactions."""
    pass

class Shuffle(pipes.Shuffle, EnvironmentFilter):
    """Shuffle interaction order."""

    def filter(self, interactions: Iterable[Interaction]) -> Sequence[Interaction]:
        first, interactions = peek_first(interactions)

        if not interactions:
            yield from []

        elif 'action' in first and 'reward' in first:
            #this is here because if it is not offpolicy evaluation can give a
            #very biased estimate when seeds are the same. To see this run.
                # import numpy as np
                # from coba import CobaRandom

                # N    = 1_000_000
                # seed = 9

                # R1 = CobaRandom(seed).randoms(N)
                # R2 = CobaRandom(seed).shuffle(R1)

                # np.corrcoef(R1,R2)

            old_seed = self._seed
            new_seed = self._seed * 3.21 if self._seed is not None else self._seed

            self._seed = new_seed
            yield from super().filter(interactions)
            self._seed = old_seed

        else:
            yield from super().filter(interactions)

class Reservoir(pipes.Reservoir, EnvironmentFilter):
    """Take n random interactions."""
    pass

class Cache(pipes.Cache, EnvironmentFilter):
    """Cache given interactions."""

    def filter(self, items: Iterable[Interaction]) -> Iterable[Interaction]:
        yield from map(methodcaller('copy'), super().filter(items))

class Scale(EnvironmentFilter):
    """Scale and shift features."""

    def __init__(self,
        shift: Union[Number,Literal["min","mean","median"]] = 0,
        scale: Union[Number,Literal["minmax","std","iqr","maxabs"]] = "minmax",
        target: Literal["context"] = "context",
        using: Optional[int] = None):
        """Instantiate a Scale filter.

        Args:
            shift: The statistic to use to shift each context feature.
            scale: The statistic to use to scale each context feature.
            target: The target data we wish to scale in the environment.
            using: The number of interactions to use when calculating the necessary statistics.
        """

        assert isinstance(shift,Number) or shift in ["min","mean","med","median"]
        assert isinstance(scale,Number) or scale in ["minmax","std","iqr","maxabs"]

        self._times = [0,0,0,0]

        self._shift  = shift
        self._scale  = scale
        self._using  = using
        self._target = target

    @property
    def params(self) -> Mapping[str, Any]:
        return {
            "shift": self._shift,
            "scale": self._scale,
            "scale_using": self._using,
        }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(Mutable().filter(interactions))

        if not interactions:
            return

        if 'context' not in first:
            yield from interactions
            return

        remaining_interactions = iter(interactions)
        fitting_interactions   = list(islice(remaining_interactions,self._using))

        first_context = first['context']

        if isinstance(first_context, primitives.Sparse) and self._target=="context" and self._shift != 0:
            raise CobaException("Shift is required to be 0 for sparse environments. Otherwise the environment will become dense.")

        is_dense_context  = isinstance(first_context, primitives.Dense)
        is_sparse_context = isinstance(first_context, primitives.Sparse)
        is_value_context  = not (is_dense_context or is_sparse_context)

        #get the values we wish to scale
        if self._target == "context" :
            if is_dense_context:
                potential_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float))]
                if len(potential_cols) == 0:
                    unscaled = []
                elif len(potential_cols) == 1:
                    unscaled = [ tuple(map(itemgetter(*potential_cols),map(itemgetter("context"),fitting_interactions))) ]
                else:
                    unscaled = list(zip(*map(itemgetter(*potential_cols),map(itemgetter("context"),fitting_interactions))))
            elif is_sparse_context:
                unscalable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float))}
                unscaled = defaultdict(list)
                for interaction in fitting_interactions:
                    context = interaction['context']
                    for k in context.keys()-unscalable_cols:
                        unscaled[k].append(context[k])
            elif is_value_context:
                unscaled = [interaction['context'] for interaction in fitting_interactions]

        #determine the scale and shift values
        if self._target == "context":
            if is_dense_context:
                shifts_scales = []
                for i,col in zip(potential_cols,unscaled):
                    shift_scale = self._get_shift_and_scale(col)
                    if shift_scale is not None:
                        shifts_scales.append((i,)+shift_scale)
            elif is_sparse_context:
                shifts_scales = {}
                for k,col in unscaled.items():
                    vals = col + [0]*(len(fitting_interactions)-len(col))
                    shift_scale = self._get_shift_and_scale(vals)
                    if shift_scale is not None:
                        shifts_scales[k] = shift_scale
            elif is_value_context:
                shifts_scales = self._get_shift_and_scale(unscaled)

        #now scale
        if not shifts_scales:
            yield from chain(fitting_interactions, remaining_interactions)
        elif self._target == "context":
            if is_dense_context:
                for interaction in chain(fitting_interactions, remaining_interactions):
                    context = interaction['context']
                    for i,shift,scale in shifts_scales:
                        context[i] = (context[i]+shift)*scale
                    yield interaction
            elif is_sparse_context:
                for interaction in chain(fitting_interactions, remaining_interactions):
                    new = interaction # Mutable copies
                    context = new['context']
                    for k in shifts_scales.keys() & context.keys():
                        (shift,scale) = shifts_scales[k]
                        context[k] = (context[k]+shift)*scale
                    yield new
            elif is_value_context:
                (shift,scale) = shifts_scales
                for interaction in chain(fitting_interactions, remaining_interactions):
                    new = interaction.copy()
                    if new['context'] is not None:
                        new['context'] = (new['context']+shift)*scale
                    yield new

    def _get_shift_and_scale(self,values) -> Tuple[float,float]:
        try:
            values = [v for v in values if v is not None]
            shift = self._shift_value(values)
            scale = self._scale_value(values,shift)

            if isnan(shift):
                #this is a trick, nan != nan so equality will tell
                #if a value is not equal with itself then it is nan.
                #Using this trick is about 2x faster than using isnan.
                not_nan_vals = list(compress(values,map(eq,values,values)))
                shift = self._shift_value(not_nan_vals)
                scale = self._scale_value(not_nan_vals,shift)

            return shift,scale
        except (TypeError,ValueError):
            return None

    def _shift_value(self, values) -> float:
        shift = self._shift
        if shift == "min":
            return -min(values)
        elif shift == "mean":
            return -sum(values)/len(values) #mean() is very slow due to calculations for precision
        elif shift == "med" or shift == "median":
            return -median(values)
        return shift

    def _scale_value(self, values, shift) -> float:
        scale     = self._scale
        scale_num = scale
        scale_den = 1

        if scale == "minmax":
            scale_num = 1
            scale_den = max(values)-min(values)
        elif scale == "std":
            scale_num = 1
            scale_den = stdev(values)
        elif scale == "iqr":
            scale_num = 1
            scale_den = iqr(values)
        elif scale == "maxabs":
            scale_num = 1
            scale_den = max([abs(v+shift) for v in values])

        return scale_num if scale_den < .000001 else scale_num/scale_den

class Impute(EnvironmentFilter):
    """Impute missing values (nan) in Interaction contexts."""

    def __init__(self,
        stat : Literal["mean","median","mode"] = "mean",
        indicator: bool = True,
        using: Optional[int] = None):
        """Impute missing data.

        Args:
            stats: The statistic to use for imputation.
            indicator: Indicates whether a binary feature should be added for missingness.
            using: The number of interactions to use to calculate imputation statistics.
        """

        assert stat in ["mean","median","mode"]

        self._stat  = stat
        self._miss  = indicator
        self._using = using
        self._times = [0,0,0,0]

    @property
    def params(self) -> Mapping[str, Any]:
        return { "impute_stat": self._stat, "impute_using": self._using, "impute_indicator":self._miss }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        first, interactions = peek_first(Mutable().filter(interactions))

        if not interactions:
            return

        if 'context' not in first:
            yield from interactions
            return

        is_dense  = isinstance(first['context'], primitives.Dense)
        is_sparse = isinstance(first['context'], primitives.Sparse)
        is_value  = not is_dense and not is_sparse

        interactions       = iter(interactions)
        using_interactions = list(islice(interactions,self._using))
        other_interactions = interactions

        start = time.time()
        if is_dense:
            if self._stat in ["mean","median"]:
                imputable_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float) or v is None)]
            else:
                imputable_cols = list(range(len(first['context'])))

        elif is_sparse:
            if self._stat in ['mean','median']:
                unimputable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float) or v is None)}
            else:
                unimputable_cols = {}

        elif is_value:
            if self._stat in ["mean","median"]:
                imputable_cols = isinstance(first['context'],(int,float)) or (first['context'] is None)
            else:
                imputable_cols = True
        self._times[0] += time.time()-start

        start = time.time()
        #get unimputed values
        if is_dense:
            if len(imputable_cols) == 0:
                unimputed = []
            elif len(imputable_cols) == 1:
                unimputed = [ tuple(map(itemgetter(*imputable_cols),map(itemgetter("context"),using_interactions))) ]
            else:
                unimputed = list(zip(*map(itemgetter(*imputable_cols),map(itemgetter("context"),using_interactions))))

        elif is_sparse:
            unimputed = defaultdict(list)
            for interaction in using_interactions:
                context = interaction['context']
                for k in context.keys()-unimputable_cols:
                    unimputed[k].append(context[k])

        elif is_value:
            unimputed = [ interaction['context'] for interaction in using_interactions ]
        self._times[1] += time.time()-start

        start = time.time()
        #calculate imputation statistics
        if is_dense:
            imputations = {}
            impute_binary = {}
            for i, col in zip(imputable_cols,unimputed):
                imputation = self._get_imputation(col)
                if imputation is not None:
                    imputations[i] = imputation
                    if self._miss and any([c is None for c in col]):
                        impute_binary[i] = len(impute_binary)

        elif is_sparse:
            imputations = {}
            impute_binary = {}
            binary_template = {}
            for k,col in unimputed.items():
                imputation = self._get_imputation(col + [0]*(len(using_interactions)-len(col)))
                if imputation is not None:
                    imputations[k] = imputation
                    if self._miss and any([c is None for c in col]):
                        impute_binary[k] = f"{k}_is_missing"
                        binary_template[f"{k}_is_missing"] = 0

        elif is_value:
            imputations = self._get_imputation(unimputed)
            impute_binary = self._miss and any([c is None for c in unimputed])
        self._times[2] += time.time()-start

        start = time.time()
        for interaction in chain(using_interactions,other_interactions):

            context = interaction['context']

            if is_dense:
                is_missing = [0]*len(impute_binary)
                for k,v in enumerate(context):
                    if v is None and k in imputations:
                        context[k] = imputations[k]
                        if k in impute_binary:
                            is_missing[impute_binary[k]] = 1
                context += is_missing

            elif is_sparse:

                is_missing = binary_template.copy()
                for k,v in context.items():
                    if v is None:
                        context[k] = imputations[k]
                        if k in impute_binary:
                            is_missing[impute_binary[k]] = 1
                context.update(is_missing)

            elif is_value:
                if impute_binary:
                    if context is None:
                        interaction["context"] = [imputations,1]
                    else:
                        interaction["context"] = [context,0]
                else:
                    if context is None:
                        interaction["context"] = imputations

            yield interaction
        self._times[3] += time.time()-start

    def _get_imputation(self,values):
        try:
            values = [v for v in values if v is not None]
            if self._stat == "mean":
                return sum(values)/len(values)
            if self._stat == "median":
                return median(values)
            if self._stat == "mode":
                return mode(values)
        except:
            return None

class Sparsify(EnvironmentFilter):
    """Ensure that features are sparse."""

    def __init__(self, context:bool = True, action:bool = False):
        """Instantiate a Sparsify filter.

        Args:
            context: If True then contexts should be made sparse otherwise leave them alone.
            action: If True then actions should be made sparse otherwise leave them alone.
        """

        self._context = context
        self._action  = action

    @property
    def params(self) -> Mapping[str, Any]:
        return { "sparse_c": self._context, "sparse_a": self._action }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first,interactions = peek_first(interactions)

        context_has_headers = 'context' in first and hasattr(first['context']   ,'headers')
        actions_has_headers = 'actions' in first and hasattr(first['actions'][0],'headers')
        action_has_headers  = 'action'  in first and hasattr(first['action' ]   ,'headers')

        for interaction in interactions:

            new = interaction.copy()

            if self._context and 'context' in new:
                new['context'] = self._make_sparse(new['context'], context_has_headers, 'context')

            if self._action and 'actions' in new:
                new['actions'] = list(map(self._make_sparse,new['actions'],repeat(actions_has_headers),repeat('action')))

            if self._action and 'action' in new:
                new['action'] = self._make_sparse(new['action'],action_has_headers,'action')

            yield new

    def _make_sparse(self, value, has_headers:bool, default_header:str) -> Optional[dict]:
        if value is None:
            return value

        if isinstance(value,primitives.Dense):
            if has_headers:
                value_list = list(value)
                return {k:value_list[i] for k,i in value.headers.items() if i < len(value_list) and value_list[i] != 0}
            else:
                return {str(k):v for k,v in enumerate(value) if v != 0 }

        if isinstance(value,primitives.Sparse):
            return value

        return {default_header:value}

class Densify(EnvironmentFilter):
    """Ensure that features are dense."""

    def __init__(self,
        n_feats: int = 400,
        method : Literal['lookup','hashing'] = 'lookup',
        context: bool = True,
        action : bool = False):
        """Instantiate a Densify filter.

        Args:
            n_feats: The number of features densified environment should have.
            method: How sparse features are turned into dense features. The hashing
                trick is more memory efficient but may have collisions. The lookup
                method is less memory efficient but guaranteed to have no collisions.
            context: If True then contexts should be made sparse otherwise leave them alone.
            action: If True then actions should be made sparse otherwise leave them alone.

            Remarks:
                The "lookup" method produces a more efficient feature representation but its memory
                usage expand to the cardinality of the sparse features. The "hashing" method utilizes
                a less efficient feature representation but does not use any memory beyond the features.
                When using the 'hashing' method `n_feats` should be set to a large number such as 2**16.
        """

        self._n_feats = n_feats
        self._context = context
        self._action  = action
        self._method  = method

        def generator():
            rng = CobaRandom(seed=1)
            while True:
                for i in rng.shuffle(range(n_feats)):
                    yield i

        def factory(g=generator()):
            return next(g)

        self._lookup = defaultdict(factory)

    @property
    def params(self) -> Mapping[str, Any]:
        return { "dense_m": self._method, "dense_n": self._n_feats, "dense_c": self._context, "dense_a": self._action }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        for interaction in interactions:

            new = interaction.copy()

            if self._context and 'context' in new:
                new['context'] = self._make_dense(new['context'])

            if self._action and 'actions' in new:
                new['actions'] = list(map(self._make_dense,new['actions']))

            if self._action and 'action' in new:
                new['action'] = self._make_dense(new['action'])

            yield new

    def _make_dense(self, value) -> Optional[dict]:

        #For this conversion we use the hashing trick. Based on many tests we go with crc32:
        #   > It is fast
        #   > It has few collisions
        #   > It is deterministic between runs

        # Hashing Algorithms We Tested:
        #   from hashlib import md5, sha1, sha256, sha3_128, sha3_256, shake_128, shake_256, blake2b
        #   from zlib import adler32, crc32
        #   from mmh3 import hash as mmh3_hash #(aka, MurmurHash3, requires pip install mmh3)
        #   from builtins import hash

        # Performance Findings:
            # hashlib algorithms took between 0.5 to 0.75 seconds to produce 1 million indexes
            # zlib and mmh3 algorithms took about .16 seconds to produce 1 million indexes
            # builtin hash algorithm took about .09 seconds to produce 1 million indexes

        # Collision Findings:
            # we used the code below to count collisions for 1,000,000 randomly generated strings:
                #import secrets
                #n=10**6
                #1-len(set([<hash-algo>(secrets.token_bytes(7)) for _ in range(n)]))/n

            # hashlib and the builtin hash had 0 collisions
            # mmh3 and crc32 had ~130 collisions with 1 million items (i.e., 0.013%)
            # adler32 had ~250k collisions (i.e., 24.00%)

        #Final decisions:
            # crc32:
            #   > It is considerably faster than the hashlib algorithms
            #   > It has equal runtime and collisions to MurmurHash3
            #   > It is deterministic across runs (unlike the built-in hash)

        if not isinstance(value,primitives.Sparse):
            return value

        else:
            if self._method == 'lookup':
                values = { self._lookup[k]:v for k,v in value.items() }
            else:
                values = { crc32(k.encode('ascii')) % self._n_feats: v for k,v in value.items() }

            return SparseDense(values, self._n_feats)

class Cycle(EnvironmentFilter):
    """Cycle reward values.

    Useful for testing a learner's response to a non-stationary shock.
    """

    def __init__(self, after:int = 0):
        """Instantiate a Cycle filter.

        Args:
            after: The number of interactions to wait before cycling reward values.
        """
        self._after = after

    @property
    def params(self) -> Mapping[str, Any]:
        return { "cycle_after": self._after }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(interactions)

        has_actions   = first and 'actions'   in first
        has_rewards   = first and 'rewards'   in first
        has_feedbacks = first and 'feedbacks' in first

        targets = []
        if has_rewards  : targets.append(['rewards'  ,callable(first['rewards'  ])])
        if has_feedbacks: targets.append(['feedbacks',callable(first['feedbacks'])])

        one_hot = lambda n,N: tuple([0]*n+[1]+[0]*(N-n-1))
        rotate  = lambda l: l[-1%n_actions:] + l[:-1%n_actions]

        n_actions      = len(first['actions']) if has_actions else 0
        is_discrete    = 0 < n_actions and n_actions < float('inf')
        onehot_actions = set(map(one_hot,range(n_actions),repeat(n_actions)))
        is_onehot      = has_actions and set(first['actions']) == onehot_actions
        is_categorical = has_actions and all(map(isinstance,first['actions'],repeat(str)))

        is_cyclable = has_actions and has_rewards and is_discrete and (is_onehot or is_categorical)

        if not is_cyclable:
            warnings.warn("Cycle only works for discrete environments without action features. It will be ignored in this case.")
            yield from interactions

        else:
            for i,old in enumerate(interactions):
                new = old.copy()
                if i >= self._after:
                    for target,is_callable in targets:
                        if not is_callable:
                            new[target] = rotate(new[target])
                        else:
                            new[target] = DiscreteReward(old['actions'],rotate(list(map(old[target],old['actions']))))
                yield new

class Flatten(EnvironmentFilter):
    """Flatten contexts and actions."""

    def __init__(self):
        self._flattener = pipes.Flatten()

    @property
    def params(self) -> Mapping[str, Any]:
        return { "flat": True }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first,interactions = peek_first(interactions)

        has_context = first and 'context' in first
        has_actions = first and 'actions' in first

        if not interactions:
            return

        if not has_context and not has_actions:
            yield from interactions
            return

        I1,I2,I3 = tee(interactions,3)

        interactions = I1

        if has_context: flat_context_iter = self._flattener.filter(i['context'] for i in I2                      )
        if has_actions: flat_actions_iter = self._flattener.filter(a            for i in I3 for a in i['actions'])

        targets = []
        if callable(first.get('rewards'))  : targets.append('rewards')
        if callable(first.get('feedbacks')): targets.append('feedbacks')

        for old in interactions:
            new = old.copy()

            if has_context:
                new['context'] = next(flat_context_iter)

            if has_actions:
                new['actions'] = list(islice(flat_actions_iter,len(old['actions'])))

                if targets and new['actions'] != old['actions']:
                    for target in targets:
                        new[target] = DiscreteReward(new['actions'],list(map(old[target],old['actions'])))

            yield new

class Binary(EnvironmentFilter):
    """Transform reward values to 1 or 0."""

    @property
    def params(self) -> Mapping[str, Any]:
        return { "binary": True }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        first, interactions = peek_first(interactions)

        is_discrete = 'actions' in first and 0 < len(first['actions']) and len(first['actions']) < float('inf')
        is_callable = callable(first['rewards'])

        if not is_discrete:
            warnings.warn("Only discrete action rewards can be made binary. ")
            yield from interactions

        else:
            for interaction in interactions:
                new     = interaction.copy()
                actions = interaction['actions']
                rewards = interaction['rewards']

                if is_callable:
                    argmax = max(actions, key=rewards)
                    new['rewards'] = BinaryReward(argmax)
                else:
                    argmax = max(range(len(actions)),key=rewards.__getitem__)
                    new['rewards'] = [0] * len(new['actions'])
                    new['rewards'][argmax] = 1

                yield new

class Sort(EnvironmentFilter):
    """Sort interactions by features."""

    def __init__(self, *keys: Union[str,int,Sequence[Union[str,int]]]) -> None:
        """Instantiate a Sort filter.

        Args:
            *keys: The context items that should be sorted on.
        """

        self._keys = list(pipes.Flatten().filter([list(keys)]))[0]

    @property
    def params(self) -> Mapping[str, Any]:
        return { "sort_keys": self._keys or '*' }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(interactions)

        if not interactions:
            return

        if first and 'context' not in first:
            yield from interactions
            return

        full_sorter = lambda interaction: tuple(interaction['context']                                 )
        list_sorter = lambda interaction: tuple(interaction['context'][key]       for key in self._keys)
        dict_sorter = lambda interaction: tuple(interaction['context'].get(key,0) for key in self._keys)

        first, interactions = peek_first(interactions)
        is_sparse           = isinstance(first['context'],primitives.Sparse)

        sorter = full_sorter if not self._keys else dict_sorter if is_sparse else list_sorter
        yield from sorted(interactions, key=sorter)

class Where(EnvironmentFilter):
    """Select for characteristics"""

    def __init__(self, *,
        n_interactions: Union[int,Tuple[Optional[int],Optional[int]]] = None,
        n_actions: Union[int,Tuple[Optional[int],Optional[int]]] = None,
        n_features: Union[int,Tuple[Optional[int],Optional[int]]] = None) -> None:
        """Instantiate a Where filter.

        Args:
            n_interactions: The minimum, maximum or exact number of interactions Environments must have.
            n_actions: The minimum, maximum or exact number of actions interactions must have.
            n_features: The minimum, maximum or exact number of features interactions must have.
        """

        self._n_interactions = n_interactions
        self._n_actions = n_actions
        self._n_features = n_features

    @property
    def params(self) -> Mapping[str, Any]:
        params = {}

        if self._n_interactions is not None:
            params["where_n_interactions"] = self._n_interactions

        if self._n_actions is not None:
            params["where_n_actions"] = self._n_actions

        if self._n_features is not None:
            params["where_n_features"] = self._n_features

        return params

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        n_int = self._n_interactions
        n_int = list(n_int) if isinstance(n_int,(list,tuple)) else [n_int]*2

        n_act = self._n_actions
        n_act = list(n_act) if isinstance(n_act,(list,tuple)) else [n_act]*2

        n_fet = self._n_features
        n_fet = list(n_fet) if isinstance(n_fet,(list,tuple)) else [n_fet]*2

        firstn = 1+next((v for v in [*n_int,0] if v is not None))
        first,interactions = peek_first(interactions, n=firstn, reduce=False)

        if not interactions:
            return

        if not self._in_min_max(len(first), *n_int):
            return

        if not self._in_min_max(self._context_len(first), *n_fet):
            return

        for interaction in interactions:
            if n_act == [None,None] or self._in_min_max(len(interaction['actions']),*n_act):
                yield interaction

    def _in_min_max(self,v:int, minv:Optional[float], maxv:Optional[float]):
        return (minv is None or minv <= v) and (maxv is None or v <= maxv)

    def _context_len(self,firstn) -> int:
        try:
            context = firstn[0].get('context')
            return try_else(lambda: len(context), 1) if context or context == 0 else 0
        except:
            return 0

class Riffle(EnvironmentFilter):
    """Riffle shuffle interactions."""

    def __init__(self, spacing: int = 3, seed=1) -> None:
        """Instantiate a Riffle filter.

        Args:
            spacing: The number of interactions from the beginning between each interaction shuffled in from the end.
            seed: The seed used to determine the location of each ending interaction when placed within its beginning space.
        """
        self._spacing = spacing
        self._seed = seed

    @property
    def params(self) -> Mapping[str, Any]:
        return {"riffle_spacing": self._spacing, "riffle_seed": self._seed}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        rng          = CobaRandom(self._seed)
        interactions = list(interactions)

        for i in range(int(len(interactions)/(self._spacing+1))):
            interactions.insert(i*self._spacing+rng.randint(0,self._spacing), interactions.pop())

        return interactions

class Noise(EnvironmentFilter):
    """Add noise to values."""

    def __init__(self,
        context: Union[Tuple[str,float,float],Callable[[float,CobaRandom], float]] = None,
        action : Union[Tuple[str,float,float],Callable[[float,CobaRandom], float]] = None,
        reward : Union[Tuple[str,float,float],Callable[[float,CobaRandom], float]] = None,
        seed   : int = 1) -> None:
        """Instantiate a Noise EnvironmentFilter.

        Args:
            context: A noise generator for context features.
            action : A noise generator for action features.
            reward : A noise generator for rewards.
            seed   : The seed initializing the random state of the noise generators.
        """

        self._args = (context,action,reward,seed)

        def make_noiser(p):
            if not p:
                return None
            elif callable(p):
                return p
            elif len(p) == 2:
                return (lambda x, rng, p=p: x+rng.gauss(*p))
            elif p[0] == 'g':
                return (lambda x, rng, p=p[1:]: x+rng.gauss(*p))
            elif p[0] == 'i':
                return (lambda x, rng, p=p[1:]: x+rng.randint(*p))

        if context is None and action is None and reward is None: context = ('g',0,1)

        self._context_noise = make_noiser(context)
        self._action_noise  = make_noiser(action)
        self._reward_noise  = make_noiser(reward)
        self._seed          = seed

    def __reduce__(self) -> tuple:
        try:
            pickle.dumps(self._args)
        except Exception:
            message = (
                "We were unable to pickle the Noise filter. This is likely due to using lambda functions for noise generation. "
                "To work around this we recommend you first define your lambda functions as a named function and then pass the "
                "named function to Noise."
            )
            raise CobaException(message)
        else:
            return (Noise, self._args)

    @property
    def params(self) -> Mapping[str, Any]:

        params = {}

        if self._context_noise: params['context_noise'] = True
        if self._action_noise : params['action_noise' ] = True
        if self._reward_noise : params['reward_noise' ] = True

        params['noise_seed'] = self._seed

        return params

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        rng = CobaRandom(self._seed)
        first,interactions = peek_first(interactions)

        if not interactions:
            return

        is_callable = callable(first.get('rewards'))

        for interaction in interactions:

            new = interaction.copy()

            if 'context' in new:
                context = new['context']
                noisy_context = self._noises(context, rng, self._context_noise)
                new['context'] = noisy_context

            if 'actions' in new:
                actions = new['actions']
                noisy_actions = [ self._noises(a, rng, self._action_noise) for a in actions ]
                new['actions'] = noisy_actions

            if 'rewards' in new:
                rewards = new['rewards']
                if is_callable: rewards = map(rewards,actions)
                noisy_rewards = [ self._noises(r, rng, self._reward_noise) for r in rewards ]
                if is_callable: noisy_rewards = DiscreteReward(noisy_actions, noisy_rewards)
                new['rewards'] = noisy_rewards

            yield new

    def _noises(self, value:Union[None,float,str,Mapping,Sequence], rng: CobaRandom, noiser: Callable[[float,CobaRandom], float]):
        if noiser is None:
            return value
        if isinstance(value, primitives.Sparse):
            #we sort so that noise generation is deterministic with respect to seed
            return { k:self._noise(v, rng, noiser) for k,v in sorted(value.items()) }
        if isinstance(value, primitives.Dense):
            return [ self._noise(v, rng, noiser) for v in value ]
        return self._noise(value, rng, noiser)

    def _noise(self, value:Union[None,float,str], rng: CobaRandom, noiser: Callable[[float,CobaRandom], float]) -> float:
        return value if not isinstance(value,(int,float)) else noiser(value, rng)

class Params(EnvironmentFilter):
    """Add params to an environment."""
    def __init__(self, params: Mapping[str, Any]) -> None:
        self._params = params

    @property
    def params(self) -> Mapping[str, Any]:
        return self._params

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        return interactions

class Grounded(EnvironmentFilter):
    """Transform simulated interactions to IGL interactions."""

    class GroundedFeedback:
        __slots__ = ('_rng','_goods','_bads','_seed','_argmax')
        def __init__(self, goods, bads, argmax, seed):
            self._rng    = None
            self._goods  = goods
            self._bads   = bads
            self._seed   = seed
            self._argmax = argmax

        @lru_cache(maxsize=None)
        def __call__(self, arg):
            if not self._rng:
                self._rng = CobaRandom(self._seed)
            if arg == self._argmax:
                return self._rng.choice(self._goods)
            else:
                return self._rng.choice(self._bads)

        def __repr__(self) -> str:
            am = self._argmax
            return f"GroundedFeedback({try_else(lambda:minimize(am),str(am))})"

    def __init__(self, n_users: int, n_normal:int, n_words:int, n_good:int, seed:int) -> None:
        """Instantiate a Grounded filter.

        Args:
            n_users: The number of users in the grounded environment.
            n_normal: The number of users with normal grounded behavior.
            n_words: The number of potential feedback words for users.
            n_good: The number of words that mean good out of the n_words.
            seed: Seed for all random values.

        Remarks:
            See `here`__ for more on interaction grounded learning.

        __ https://proceedings.mlr.press/v139/xie21e.html
        """
        self._n_users  = self._cast_int(n_users , "n_users" )
        self._n_normal = self._cast_int(n_normal, "n_normal")
        self._n_words  = self._cast_int(n_words , "n_words" )
        self._n_good   = self._cast_int(n_good  , "n_good"  )
        self._seed     = seed

        if n_normal > n_users:
            raise CobaException("Igl conversion can't have more normal users (n_normal) than total users (n_users).")

        if n_good > n_words:
            raise CobaException("Igl conversion can't have more good words (n_good) than total words (n_words).")

        self.userids   = list(range(self._n_users))
        self.normalids = self.userids[:self._n_normal]
        self.wordids   = list(range(self._n_words))
        self.goodwords = self.wordids[:self._n_good]
        self.badwords  = self.wordids[self._n_good:]

    @property
    def params(self) -> Mapping[str, Any]:
        return {
            "n_users"  : self._n_users,
            "n_normal" : self._n_normal,
            "n_good"   : self._n_good,
            "n_words"  : self._n_words,
            "igl_seed" : self._seed
        }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        rng = CobaRandom(self._seed)

        #we make it a set for faster contains checks
        normalids       = set(self.normalids)
        normal          = [u in normalids for u in self.userids]
        userid_isnormal = list(zip(self.userids,normal))

        first,interactions = peek_first(interactions)

        if not interactions:
            return

        is_callable = callable(first['rewards'])

        if is_callable:
            is_binary_rwd = set(map(first['rewards'], first['actions'])) == {0,1}
        else:
            is_binary_rwd = set(first['rewards']) == {0,1}

        goods = [(g,) for g in self.goodwords]
        bads  = [(b,) for b in self.badwords ]

        for seed, interaction in enumerate(interactions, self._seed):

            userid,normal   = rng.choice(userid_isnormal)
            actions,rewards = interaction['actions'],interaction['rewards']

            new = interaction.copy()

            if is_callable:
                argmax = max(actions,key=rewards)
            else:
                argmax = actions[max(range(len(actions)),key=rewards.__getitem__)]

            if not is_binary_rwd:
                new['rewards'] = BinaryReward(argmax)

            if normal:
                new['feedbacks'] = Grounded.GroundedFeedback(goods,bads,argmax,seed)
            else:
                new['feedbacks'] = Grounded.GroundedFeedback(bads,goods,argmax,seed)

            new['userid'  ] = userid
            new['isnormal'] = normal

            yield new

    def _cast_int(self, value:Union[float,int], value_name:str) -> int:
        if isinstance(value, int): return value
        if isinstance(value, float) and value.is_integer(): return int(value)
        raise CobaException(f"{value_name} must be a whole number and not {value}.")

class Repr(EnvironmentFilter):
    """Change representation of categorical data."""

    def __init__(self,
        categorical_context:Literal["onehot","onehot_tuple","string"] = None,
        categorical_actions:Literal["onehot","onehot_tuple","string"] = None) -> None:
        """Instantiate a Repr filter.

        Args:
            cat_context: How to represent categorical data in contexts.
            cat_actions: How to represent categorical data in actions.
        """
        self._cat_context = categorical_context
        self._cat_actions = categorical_actions

    @property
    def params(self) -> Mapping[str, Any]:
        return {"categoricals_in_context": self._cat_context, "categoricals_in_actions": self._cat_actions}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        firsts, interactions = peek_first(interactions,n=2)

        if not interactions: return []

        first = firsts[0]

        has_context   = 'context'   in first
        has_actions   = 'actions'   in first and bool(first['actions'])
        has_action    = 'action'    in first
        has_rewards   = 'rewards'   in first
        has_feedbacks = 'feedbacks' in first

        n_tee = 1 + bool(self._cat_context and has_context) + \
                    bool(self._cat_actions and has_actions) + \
                    bool(self._cat_actions and has_action)

        tees = iter([interactions]) if n_tee == 1 else iter(tee(interactions, n_tee))

        if not (self._cat_context and has_context):
            cat_context_iter = None
        else:
            cat_context_iter = pipes.EncodeCatRows(self._cat_context).filter(i['context'] for i in next(tees))

        if not (self._cat_actions and has_actions):
            cat_actions_iter = None
        else:
            rows = (i['actions'] for i in next(tees))

            if len(firsts) == 2 and firsts[0]['actions'] == firsts[1]['actions']:
                def yield_prev_action_on_repeat():
                    prev_row   = None
                    prev_yield = None

                    encoder = pipes.EncodeCatRows(self._cat_actions)

                    for row in rows:
                        if row != prev_row:
                            prev_row = row
                            prev_yield = list(encoder.filter(row))
                        yield prev_yield

                cat_actions_iter = yield_prev_action_on_repeat()
            else:
                def yield_action_lists():
                    r1,r2 = tee(rows,2)
                    actionitr = pipes.EncodeCatRows(self._cat_actions).filter(chain.from_iterable(r2))
                    for row in r1:
                        yield list(islice(actionitr,len(row)))
                cat_actions_iter = yield_action_lists()

        if not (self._cat_actions and has_action):
            cat_action_iter = None
        else:
            cat_action_iter = pipes.EncodeCatRows(self._cat_context).filter(i['action'] for i in next(tees))

        reward_targets = []
        if has_rewards   and callable(first['rewards'])  : reward_targets.append('rewards')
        if has_feedbacks and callable(first['feedbacks']): reward_targets.append('feedbacks')

        for old in next(tees):

            new = old.copy()

            if cat_context_iter:
                new['context'] = next(cat_context_iter)

            if cat_actions_iter:
                new['actions'] = next(cat_actions_iter)

            if cat_action_iter:
                new['action'] = next(cat_action_iter)

            actions_changed = cat_actions_iter and new['actions'] != old['actions']

            if actions_changed:
                for target in reward_targets:
                    if isinstance(old[target],BinaryReward):
                        new_argmax = new['actions'][old['actions'].index(old[target]._argmax)]
                        new[target] = BinaryReward(new_argmax,old[target]._value)
                    elif isinstance(old[target],DiscreteReward):
                        new[target] = DiscreteReward(new['actions'],old[target].rewards)
                    else:
                        new[target] = DiscreteReward(new['actions'],list(map(old[target],old['actions'])))
            yield new

class Batch(EnvironmentFilter):
    """Batch interactions."""
    class List(list):
        is_batch: bool = True

    class Callable(list):
        is_batch: bool = True
        def __call__(self, args: Any) -> Any:
            ndim = getattr(args,'ndim',-1)
            outs = list(map(lambda f,a: f(a), self, args))

            if ndim == -1:
                return Batch.List(outs)
            else:
                try:
                    t  = torch # type: ignore
                    bt = Batch.Tensor
                except:
                    t  = globals()['torch'] = __import__('torch')
                    bt = Batch._make_batch_tensor()
                return bt(t.stack(outs))

    @classmethod
    def _make_batch_tensor(cls):
        try:
            return cls.Tensor
        except:
            import torch
            class BatchTensor(torch.Tensor):
                is_batch: bool = True
            cls.Tensor = BatchTensor
            return cls.Tensor

    def __init__(self, batch_size: Optional[int], batch_type: Literal['list','torch'] = 'list') -> None:
        """Instantiate a Batch filter.

        Args:
            batch_size: The number of interactions in a batched interaction.
            batch_type: The type of batch for interaction values.
        """

        self._batch_size = batch_size or None
        self._batch_type = batch_type

        if batch_type == 'torch': PackageChecker.torch('Batch')

    @property
    def params(self) -> Mapping[str,Any]:
        return {'batch_size': self._batch_size, 'batch_type':self._batch_type}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first,interactions = peek_first(interactions)

        if not interactions:
            yield from []

        elif not self._batch_size:
            yield from interactions

        else:
            callable_keys = [k for k,v in first.items() if callable(v)]
            if self._batch_type == 'torch': self._make_batch_tensor()

            batch_data_type = Batch.List if self._batch_type == 'list' else Batch.Tensor
            batch_call_type = Batch.Callable

            for batch in self._batched(interactions, self._batch_size):

                new = {}

                for key in first.keys() & callable_keys:
                    new[key] = batch_call_type(map(itemgetter(key),batch))

                for key in first.keys() - callable_keys:
                    values = list(map(itemgetter(key),batch))
                    try:
                        new[key] = batch_data_type(values)
                    except:
                        new[key] = Batch.List(values)

                yield new

    def _batched(self, iterable, n):
        #Batch data into lists of length n.
        #The last batch may be shorter.
        #Taken from python itertools recipes.
        it = iter(iterable)
        batch = list(islice(it, n))
        while (batch):
            yield batch
            batch = list(islice(it, n))

class Unbatch(EnvironmentFilter):
    """Unbatch interactions."""

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        first, interactions = peek_first(interactions)

        if not interactions: return []

        batched_keys = [k for k,v in first.items() if is_batch(v) ]

        if batched_keys:
            yield from self._unbatch(interactions, batched_keys)
        else:
            yield from interactions

    def _unbatch(self, interactions: Iterable[Interaction], batched_keys:Sequence[str]):
        for interaction in interactions:
            batch_size = len(interaction[batched_keys[0]])
            for i in range(batch_size):
                new = {}
                for k in interaction:
                    try:
                        new[k] = interaction[k][i]
                    except:
                        new[k] = interaction[k]
                yield new

class BatchSafe(EnvironmentFilter):
    """Make a filter agnostic to batching."""

    def __init__(self, filter: EnvironmentFilter) -> None:
        """Instantiate a BatchSafe filter.

        Args:
            filter: The filter to make agnostic to batching.
        """
        self._filter = filter

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        first, interactions = peek_first(interactions)
        if not interactions: return []

        first_val  = next(iter(first.values()))
        batch_size = len(first_val) if is_batch(first_val) else None

        pipe = Pipes.join(Unbatch(),self._filter,Batch(batch_size)) if batch_size else self._filter

        yield from pipe.filter(interactions)

    def __str__(self):
        return f"BatchSafe({self._filter})"

class Harden(EnvironmentFilter):
    """Make all lazy data concrete.

    Remarks:
        Convert lazily evaluated data to concrete types.
    """

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        first, interactions = peek_first(interactions)

        if not interactions: return []

        first_has_context = 'context' in first
        first_has_actions = 'actions' in first and first['actions']
        first_has_action  = 'action'  in first

        if first_has_context:
            is_dense_context  = isinstance(first['context'],primitives.Dense)
            is_sparse_context = isinstance(first['context'],primitives.Sparse)
        if first_has_actions:
            is_dense_actions  = isinstance(first['actions'][0],primitives.Dense)
            is_sparse_actions = isinstance(first['actions'][0],primitives.Sparse)
        if first_has_action:
            is_dense_action  = isinstance(first['action'],primitives.Dense)
            is_sparse_action = isinstance(first['action'],primitives.Sparse)

        context_materialized = not first_has_context or primitives.is_materialized(first['context'])
        actions_materialized = not first_has_actions or primitives.is_materialized(first['actions'][0])
        action_materialized  = not first_has_action  or primitives.is_materialized(first['action'])

        for interaction in interactions:

            new = interaction.copy()

            if not context_materialized and is_dense_context:
                new['context'] = list(new['context'])
            elif not context_materialized and is_sparse_context:
                new['context'] = new['context'].copy()

            if not actions_materialized and is_dense_actions:
                new['actions'] = list(map(list,new['actions']))
            elif not actions_materialized and is_sparse_actions:
                new['actions'] = list(map(methodcaller('copy'),new['actions']))

            if not action_materialized and is_dense_action:
                new['action'] = list(new['action'])
            elif not action_materialized and is_sparse_action:
                new['action'] = new['action'].copy()

            yield new

class Chunk(EnvironmentFilter):
    """Indicate where to chunk.

    Remarks:
        Functionally this filter does nothing. This filter
        only has an effect when environments are passed to
        Experiment. In that case, this is useful if an early
        part of an Environments pipeline takes considerable
        time to evaluate (e.g., `logged`). Placing a chunk
        after a long-running filter means that the filter
        will only be executed once. After that, the results
        will be resued by all pipes downstream of the chunk.
    """

    def filter(self, items: Iterable[Interaction]) -> Iterable[Interaction]:
        return items

class Logged(EnvironmentFilter):
    """Transform simulated interactions to logged interactions."""

    def __init__(self, learner: Learner, seed: Optional[float] = 1.23) -> None:
        """Instantiate a Logged filter.

        Args:
            learners: The learners that will be used as the logging policy.
                An environment will be created for every learner provided.
            seed: The seed for used for all random number generation.

        Remarks:
            Adds 'action', 'reward', and 'probability' to interactions with 'context',
            'actions', and 'rewards'.
        """
        self._learner = learner
        self._seed    = seed

    @property
    def params(self) -> Mapping[str, Any]:
        lrn_params = SafeLearner(self._learner).params
        return { **lrn_params, "learner": lrn_params, "logged":True, "log_seed":self._seed}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(interactions)

        if not interactions:
            return

        if 'actions' not in first or 'rewards' not in first:
            raise CobaException("We were unable to create a logged representation of the interaction.")

        #Avoid circular dependency
        from coba.evaluators import SequentialCB

        seed = self._seed if self._seed is not None else CobaRandom().random()

        I1,I2 = tee(interactions,2)
        interactions = Unbatch().filter(I1)

        env = type("",(),{'read':lambda : I2})
        lrn = copy.deepcopy(self._learner)
        evaluator = SequentialCB(record=['action','reward','probability'],seed=seed)

        for interaction, log in zip(interactions,evaluator.evaluate(env,lrn)):
            out = interaction.copy()
            out.update(log)
            yield out

class Mutable(EnvironmentFilter):
    """Make all feature containers mutable (i.e., list or dict)."""

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(interactions)

        if not interactions:
            return

        if 'context' not in first:
            yield from interactions
            return

        first_is_dense   = isinstance(first['context'], primitives.Dense)
        first_is_sparse  = isinstance(first['context'], primitives.Sparse)
        first_is_value   = not (first_is_dense or first_is_sparse)
        first_is_mutable = isinstance(first['context'], (list,dict,SparseDense))

        if first_is_mutable:
            for interaction in interactions:
                new = interaction.copy()
                new['context'] = new['context'].copy()
                yield new

        elif first_is_dense:
            for interaction in interactions:
                new = interaction.copy()
                new['context'] = list(new['context'])
                yield new

        elif first_is_sparse:
            for interaction in interactions:
                new = interaction.copy()
                new['context'] = dict(new['context'].items())
                yield new

        elif first_is_value:
            for interaction in interactions:
                yield interaction.copy()

class OpeRewards(EnvironmentFilter):
    """Transform logged interactions to simulated interactions."""

    def __init__(self, rwd_type:Literal['IPS','DM','DR']=None, target:str = 'rewards', features=[1,'x','a','xa','xxa']):
        """Instantiate an OpeRewards filter.

        Args:
            rewards_type: How to estimate the rewards function from the logged
            data (i.e., inverse propensity score, direct method, or doubly robust).

        Remarks:
            Adds 'rewards' to interactions with 'context', 'action', 'reward',
            and 'probability'.
        """
        if rwd_type in ['DM','DR']:
            PackageChecker.vowpalwabbit(f"{rwd_type} OpeRewards")
        self._rwd_type = rwd_type
        self._target   = target
        self._features = features

    @property
    def params(self) -> Mapping[str, Any]:
        return {'ope_reward': self._rwd_type or 'None'}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        rwd_type = self._rwd_type

        if rwd_type is None:
            yield from interactions

        elif rwd_type == "IPS":
            for interaction in interactions:
                interaction = interaction.copy()
                interaction[self._target] = BinaryReward(interaction['action'], interaction['reward']/(interaction.get('probability') or 1))
                yield interaction

        elif rwd_type in ["DM","DR"]:
            from coba.learners.vowpal import VowpalMediator

            class DMReward:
                def __init__(self, vw:VowpalMediator, context:Any) -> None:
                    self._vw = vw
                    self._x  = context

                def __call__(self, action) -> Any:
                    feats   = {"x":self._x, 'a':action} if self._x is not None else {'a':action}
                    return vw.predict(vw.make_example(feats))

            class DRReward:
                def __init__(self,vw:VowpalMediator, context:Any,  off_action, off_reward, off_prob) -> None:
                    self._vw         = vw
                    self._x          = context
                    self._off_action = off_action
                    self._off_reward = off_reward
                    self._off_prob   = (off_prob or 1)

                def __call__(self, action) -> Any:
                    feats   = {"x":self._x, 'a':action} if self._x is not None else {'a':action}
                    dm_part = vw.predict(vw.make_example(feats))
                    dr_part = (self._off_reward-dm_part)/self._off_prob if action == self._off_action else 0
                    return dr_part + dm_part

            rng = CobaRandom(1)
            vw = VowpalMediator()

            # When using cb_adf to estimate rewards for actions instead of the
            # regression we would receive rewards far outside of the expected boundaries.
            #vw.init_learner("--cb_adf --interactions xa --interactions xxa --quiet ", label_type=4)

            args = "--quiet"
            if 'x' not in self._features: args = '--ignore_linear x ' + args
            if 'a' not in self._features: args = '--ignore_linear a ' + args
            if 1 not in self._features  : args = '--noconstant ' + args
            for f in self._features:
                if f not in ['x','a',1]:
                    args = f'--interactions {f} ' + args

            vw.init_learner(args, label_type=1)

            #so that we don't run forever
            #when interactions is re-iterable
            interactions = iter(interactions)

            while True:

                #Batch Shuffling reduces the correlation between
                #the reward functions learned by this VW learner
                #and potential downstream VW learners. If this
                #correlation is strong, we will over estimate
                #the downstream VW learner's performance.
                L = list(islice(interactions,4000))

                if not L: break

                for interaction in rng.shuffle(L):
                    if 'context' in interaction:
                        features = {"x": interaction['context'], "a": interaction['action']}
                    else:
                        features = {"a": interaction['action']}
                    label    = f"{interaction['reward']} {1/(interaction.get('probability') or 1)}"
                    vw.learn(vw.make_example(features,label))

                for interaction in L:
                    new = interaction.copy()

                    if rwd_type=="DR":
                        new[self._target] = DRReward(vw,new.get('context'),new['action'],new['reward'],new['probability'])
                    else:
                        new[self._target] = DMReward(vw,new.get('context'))

                    yield new

class EmptyCheck(EnvironmentFilter):

    def __init__(self) -> None:
        self._isempty = None

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        if self._isempty is None:
            interactions = peek_first(interactions)[1]
            self._isempty = interactions == []
            if not interactions:
                CobaContext.logger.log("An environment had nothing to evaluate (this is often due to having too few interactions).")

        if not self._isempty:
            yield from interactions

class Finalize(EnvironmentFilter):
    """Final preparation for built-in Evaluators.

    Remarks:
        This filter does two things:
            1. Materializes all lazy data
            2. Turns categorical types into one-hot encodings
    """

    def __init__(self):
        self._emptycheker = EmptyCheck()

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        first,interactions = peek_first(self._emptycheker.filter(interactions))

        rwds_is_list  = first and isinstance(first.get('rewards')  ,list)
        fbks_is_list  = first and isinstance(first.get('feedbacks'),list)

        interactions = Pipes.join(Harden(),Repr("onehot","onehot")).filter(interactions)

        if not (rwds_is_list or fbks_is_list):
            yield from interactions
        else:
            for interaction in interactions:
                new = interaction.copy()
                if rwds_is_list: new['rewards'  ] = DiscreteReward(new['actions'],new['rewards'  ])
                if fbks_is_list: new['feedbacks'] = DiscreteReward(new['actions'],new['feedbacks'])
                yield new



from math import hypot, erf, sqrt
from statistics import fmean
from sys import version_info
from operator import mul, sub
from bisect import bisect_left
from itertools import repeat, accumulate, compress, chain
from typing import Sequence, Tuple, Union, Optional

def iqr(values: Sequence[float]) -> float:

    if len(values) <= 1: return 0.

    values = sorted(values)

    p25,p75 = percentile(values, [0.25,0.75])

    return p75-p25

def percentile(values: Sequence[float], percentiles: Union[float,Sequence[float]], weights: Sequence[float] = None, sort: bool = True) -> Union[float, Tuple[float,...]]:

    if len(values) == 1:
        if isinstance(percentiles,(int,float)):
            return values[0]
        else:
            return list(values)*len(percentiles)

    def _percentile(values: Sequence[float], weights:Optional[Sequence[float]], percentile: float) -> float:

        assert 0 <= percentile and percentile <= 1, "Percentile must be between 0 and 1 inclusive."

        if percentile == 0:
            return values[0]

        if percentile == 1:
            return values[-1]

        if weights:
            R = bisect_left(weights,percentile)
            L = R-1
            LP = (weights[R]-percentile)/(weights[R]-weights[L])

            return LP*values[L] + (1-LP)*values[R]
        else:
            i = percentile*(len(values)-1)
            I = int(i)

            if i == I:
                return values[I]
            else:
                w = (i-I)
                return (1-w)*values[I] + w*values[I+1]

    if sort:
        if weights:
            values, weights = zip(*sorted(zip(values,weights)))
        else:
            values = sorted(values)

    if weights:
        if any(not w for w in weights):
            values = list(compress(values,weights))
        weight_sum = sum(weights[1:])
        weights    = [w/weight_sum for w in accumulate(chain([0],compress(weights[1:],weights)))]

    if isinstance(percentiles,(float,int)):
        return _percentile(values, weights, percentiles)
    else:
        return tuple([_percentile(values, weights, p) for p in percentiles ])

def phi(x: float) -> float:
    'Cumulative distribution function for the standard normal distribution'
    return (1.0 + erf(x / sqrt(2.0))) / 2.0

def mean(sample: Sequence[float]) -> float:
    return fmean(sample)

def var(sample: Sequence[float]) -> float:
    n = len(sample)

    if n == 1: return float('nan')

    if version_info >= (3,8): #pragma: no cover
        #In Python 3.8 hypot was extended to support any number of items
        #this provides a fast/safe way to calculate the sum of squares
        #and so we revert to the one-pass method and trust in hypot
        E_s2 = hypot(*sample)**2
        E_s = sum(sample)
        return (E_s2-E_s*E_s/n)/(n-1)
    else: #pragma: no cover
        #using the corrected two pass algo as recommended by
        #https://cpsc.yale.edu/sites/default/files/files/tr222.pdf
        #I've optimized this as much as I think is possible in python
        diffs = tuple(map(sub,sample,repeat(sum(sample)/n)))
        return sum(map(mul,diffs,diffs))/(n-1)

def stdev(sample: Sequence[float]) -> float:
    return var(sample)**(1/2)

class OnlineVariance:
    """Calculate sample variance in an online fashion.

    Remarks:
        This algorithm is known as Welford's algorithm and the implementation below
        is a modified version of the Python algorithm created by Wikepedia contributors (2020).

    References:
        Wikipedia contributors. (2020, July 6). Algorithms for calculating variance. In Wikipedia, The
        Free Encyclopedia. Retrieved 18:00, July 24, 2020, from
        https://en.wikipedia.org/w/index.php?title=Algorithms_for_calculating_variance&oldid=966329915
    """

    def __init__(self) -> None:
        """Instatiate an OnlineVariance calcualator."""
        self._count    = 0.
        self._mean     = 0.
        self._M2       = 0.
        self._variance = float("nan")

    @property
    def variance(self) -> float:
        """The variance of all given updates."""
        return self._variance

    def update(self, value: float) -> None:
        """Update the current variance with the given value."""
        (count,mean,M2) = (self._count, self._mean, self._M2)

        count   += 1
        delta   = value - mean
        mean   += delta / count
        delta2  = value - mean
        M2     += delta * delta2

        (self._count, self._mean, self._M2) = (count, mean, M2)

        if count > 1:
            self._variance = M2 / (count - 1)

class OnlineMean:
    """Calculate mean in an online fashion."""

    def __init__(self):
        self._n = 0
        self._mean = float('nan')

    @property
    def mean(self) -> float:
        """The mean of all given updates."""
        return self._mean

    def update(self, value:float) -> None:
        """Update the current mean with the given value."""
        self._n += 1
        alpha = 1/self._n
        self._mean = value if alpha == 1 else (1 - alpha) * self._mean + alpha * value



from itertools import chain
from typing import Any, Iterable, Union, Sequence, overload, Dict, Literal, Tuple

from coba.pipes import Pipes, IterableSource, LabelRows, Reservoir, UrlSource, CsvReader
from coba.pipes import CsvReader, ArffReader, LibsvmReader, ManikReader

from coba.utilities  import peek_first
from coba.primitives import Categorical, Source, Environment, Dense, Sparse
from coba.primitives import SimulatedInteraction
from coba.primitives import L1Reward, BinaryReward, HammingReward

class CsvSource(Source[Iterable[Dense]]):
    """Load a csv dataset.

    This is primarily used by SupervisedSimulation to create Environments for Experiments.
    """

    def __init__(self, source: Union[str,Source[Iterable[str]]], has_header:bool=False, **dialect) -> None:
        """Instantiate a CsvSource.

        Args:
            source: The data source. Accepts either a string representing the source location or another Source.
            has_header: Indicates if the CSV files has a header row.
        """
        source = UrlSource(source) if isinstance(source,str) else source
        reader = CsvReader(has_header, **dialect)
        self._source = Pipes.join(source, reader)

    def read(self) -> Iterable[Dense]:
        """Read and parse the csv source."""
        return self._source.read()

    @property
    def params(self) -> Dict[str, Any]:
        """Parameters describing the csv source."""
        return self._source.params

    def __str__(self) -> str:
        return str(self._source)

class ArffSource(Source[Union[Iterable[Dense], Iterable[Sparse]]]):
    """Load an arff dataset.

    Remarks:
        https://waikato.github.io/weka-wiki/formats_and_processing/arff_stable/
    """

    def __init__(self,source: Union[str,Source[Iterable[str]]]) -> None:
        """Instantiate an ArffSource.

        Args:
            source: The data source. Accepts either a string representing the source location or another Source.
        """
        source       = UrlSource(source) if isinstance(source,str) else source
        self._source = Pipes.join(source, ArffReader())

    def read(self) -> Union[Iterable[Dense], Iterable[Sparse]]:
        """Read and parse the arff source."""
        return self._source.read()

    @property
    def params(self) -> Dict[str, Any]:
        """Parameters describing the arff source."""
        return self._source.params

    def __str__(self) -> str:
        return str(self._source)

class LibSvmSource(Source[Iterable[Sparse]]):
    """Load a libsvm dataset.

    Remarks:
        https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/multiclass.html
    """

    def __init__(self, source: Union[str,Source[Iterable[str]]]) -> None:
        """Instantiate a LibsvmSource.

        Args:
            source: The data source. Accepts either a string representing the source location or another Source.
        """
        source       = UrlSource(source) if isinstance(source,str) else source
        self._source = Pipes.join(source, LibsvmReader())

    def read(self) -> Iterable[Sparse]:
        """Read and parse the libsvm source."""
        return self._source.read()

    @property
    def params(self) -> Dict[str, Any]:
        """Parameters describing the libsvm source."""
        return self._source.params

    def __str__(self) -> str:
        return str(self._source)

class ManikSource(Source[Iterable[Sparse]]):
    """Load a manik dataset.

    Remarks:
        http://manikvarma.org/downloads/XC/XMLRepository.html

    """

    def __init__(self, source: Union[str,Source[Iterable[str]]]) -> None:
        """Instantiate a ManikSource.

        Args:
            source: The data source. Accepts either a string representing the source location or another Source.
        """
        source       = UrlSource(source) if isinstance(source,str) else source
        self._source = Pipes.join(source, ManikReader())

    def read(self) -> Iterable[Sparse]:
        """Read and parse the manik source."""
        return self._source.read()

    @property
    def params(self) -> Dict[str, Any]:
        """Parameters describing the manik source."""
        return self._source.params

    def __str__(self) -> str:
        return str(self._source)

class SupervisedSimulation(Environment):
    """A contextual bandit environment created from supervised data."""

    @overload
    def __init__(self,
        source: Source[Union[Iterable[Dense], Iterable[Sparse], Iterable[Tuple[Any,Any]]]] = None,
        label_col: Union[int,str] = None,
        label_type: Literal["c","r","m"] = None,
        take: int = None) -> None:
        """Instantiate a SupervisedSimulation.

        Args:
            source: A source object that reads the supervised data.
            label_col: The header name or index which identifies the label feature in each example. If
                label_col is None the source must return an iterable of tuple pairs where the first item
                are the features and the second item is the label.
            label_type: Indicates whether the label column is a classification or regression value. If an explicit
                label_type is not provided then the label_type will be inferred based on the data source.
            take: The number of random examples you'd like to draw from the given data set for the environment.
        """
        ...

    @overload
    def __init__(self,
        X: Sequence[Any],
        Y: Sequence[Any],
        label_type: Literal["c","r","m"]) -> None:
        """Instantiate a SupervisedSimulation.

        Args:
            X: A sequence of example features that will be used to create interaction contexts in the simulation.
            Y: A sequence of supervised labels that will be used to construct actions and rewards in the simulation.
            label_type: Indicates whether the label column is a classification or regression value.
        """
        ...

    def __init__(self, *args, **kwargs) -> None:
        """Instantiate a SupervisedSimulation."""

        if 'source' in kwargs or (args and hasattr(args[0], 'read')):
            source     = args[0] if len(args) > 0 else kwargs['source']
            label_col  = args[1] if len(args) > 1 else kwargs.get("label_col", None)
            label_type = args[2] if len(args) > 2 else kwargs.get("label_type", None)
            take       = args[3] if len(args) > 3 else kwargs.get("take", None)
            if take      is not None: source = Pipes.join(source, Reservoir(take))
            if label_col is not None: source = Pipes.join(source, LabelRows(label_col,label_type))
            params = source.params

        else:
            X          = args[0]
            Y          = args[1]
            label_type = args[2] if len(args) > 2 else kwargs.get("label_type", None)
            params     = {"source": "[X,Y]"}
            source     = IterableSource(zip(X,Y))

        self._label_type = label_type
        self._source     = source
        self._params     = {**params, "label_type": self._label_type, "env_type": "SupervisedSimulation" }

    @property
    def params(self) -> Dict[str,Any]:
        return self._params

    def read(self) -> Iterable[SimulatedInteraction]:

        first, rows = peek_first(self._source.read())
        first_row_type = 0 if hasattr(first,'label') else 1

        if not rows: return

        first_label      = first.label if first_row_type == 0 else first[1]
        first_label_type = first.tipe  if first_row_type == 0 else None

        if first_label_type is None:
            label_type = self._label_type or ("r" if isinstance(first_label, (int,float)) else "c")
        else:
            label_type = self._label_type or first_label_type

        label_type = label_type.lower()

        if label_type == "r":
            actions = []
            reward  = L1Reward
            self._params['n_actions'] = float('inf')

        elif label_type == "c" and isinstance(first_label, Categorical):
            #Handling the categoricals separately allows for a performance optimization
            #since we can use the Categorical's as_int property rather than action_indexes
            actions = [ Categorical(l,first_label.levels) for l in first_label.levels ]
            reward  = BinaryReward
            self._params['n_actions'] = len(actions)

        else:
            #we need to know all labels in the dataset to determine actions
            rows = list(rows)
            lbls = [r.label for r in rows] if first_row_type == 0 else [r[1] for r in rows]
            if label_type == "m":
                actions = sorted(set(list(chain(*lbls))))
                reward = HammingReward
            else:
                delist  = lambda l: l[0] if isinstance(l,list) else l
                actions = sorted(set(map(delist,lbls)))
                reward  = lambda l: BinaryReward(delist(l))
            self._params['n_actions'] = len(actions)

        if first_row_type == 0:
            for row in rows:
                yield {'context':row.feats,'actions':actions,'rewards':reward(row.label)}
        else:
            for row in rows:
                yield {'context':row[0],'actions':actions,'rewards':reward(row[1])}



from math import isclose
from collections import abc
from typing import Union, Tuple, Mapping, Iterable, Literal, Callable, Optional, Any

from coba.context import CobaContext
from coba.random import CobaRandom
from coba.utilities import try_else
from coba.exceptions import CobaException
from coba.primitives import Pred, Actions, Action, Context, Prob, Kwargs
from coba.primitives import Environment, Evaluator, Learner, Interaction, is_batch

class SafeEnvironment(Environment):
    """A wrapper for environment-likes that guarantees interface consistency."""

    def __init__(self, environment: Environment) -> None:
        """Instantiate a SafeEnvironment.

        Args:
            environment: The environment we wish to make sure has the expected interface
        """
        self.env = environment if not isinstance(environment, SafeEnvironment) else environment.env

    @property
    def params(self) -> Mapping[str, Any]:
        try:
            params = self.env.params
        except AttributeError:
            params = {}

        if "env_type" not in params:
            try:
                params["env_type"] = self.env[0].__class__.__name__
            except:
                params["env_type"] = self.env.__class__.__name__

        return params

    def read(self) -> Iterable[Interaction]:
        return self.env.read()

    def __str__(self) -> str:
        params = dict(self.params)
        tipe   = params.pop("env_type")

        if len(params) > 0:
            return f"{tipe}({','.join(f'{k}={v}' for k,v in params.items())})"
        else:
            return tipe

class SafeLearner(Learner):
    """A wrapper for learner-likes that guarantees interface consistency."""

    @staticmethod
    def first_row(pred: Pred, batch_order:Literal['not','row','col'], has_kwargs:bool) -> Tuple[bool,Pred]:
        if batch_order == 'col':
            if has_kwargs and isinstance(pred[0],dict):
                pred = pred[0]
            if isinstance(pred,dict):
                return {k:v[0] for k,v in pred.items()}
            else:
                row = [p[0] for p in (pred[:-1] if has_kwargs else pred) ]
                return row[0] if len(row) == 1 else row
        else:
            if batch_order == 'row': pred = pred[0]
            return (pred[0] if len(pred)==2 else pred[:-1]) if has_kwargs else pred

    @staticmethod
    def has_kwargs(pred: Pred, batch_order:Literal['not','row','col']) -> bool:
        try:
            return isinstance(pred[-1] if batch_order != 'row' else pred[0][-1], abc.Mapping)
        except:
            return False

    @staticmethod
    def batch_order(predictor, pred: Pred, context, actions, method=None) -> Literal['not','col','row']:
        if method == 2: return 'row'
        if not is_batch(actions) and not is_batch(context): return 'not'

        no_len         = lambda item: not hasattr(item,'__len__')
        is_all_dicts   = all(isinstance(p,dict) for p in pred)
        is_dict_col    = isinstance(pred,dict)
        is_dict_col_kw = is_all_dicts and pred[0].keys() != pred[-1].keys() and len(pred)==2
        is_dict_row    = is_all_dicts and pred[0].keys() == pred[-1].keys()

        if is_dict_col or is_dict_col_kw : return 'col'
        if is_dict_row or no_len(pred[0]): return 'row'
        #if isinstance(pred[0],dict)

        n_rows = len(actions)
        n_dim1 = len(pred)
        n_dim2 = len(pred[0])

        if n_dim1 == n_dim2:
            #The major order of pred is not determinable. So we
            #now do a small "test" to determine the major order.
            #n_cols will always be >= 2 so we know we can distinguish
            class Batch(list): is_batch=True
            pred   = predictor(Batch([context[0]]),Batch([actions[0]]))
            n_rows = 1

        return 'row' if len(pred) == n_rows else 'col'

    @staticmethod
    def pred_format(std_pred:Pred, actions:Actions, og_pred:Pred = None):
        unclear_format = CobaException("We were unable to determine the prediction format from the "
        f"given value: {og_pred}. To work around this you can provide explicit format information by "
        "returnning a dict wrapper: {'pmf':<pred>}, {'action':<pred>}, or {'action_prob':<pred>}.")
        no_act_two  = CobaException("We were given a two item pred without actions. We cannot tell "
        "if this is an action or action_prob. Please use explicit hints to let us know by returning "
        "either {'action':<pred>} or {'action_prob':<pred>}.")
        no_act_pmf  = CobaException("We were given a PMF but there are no actions to choose from.")
        bad_len_pmf = CobaException("We were given a PMF whose length did match len(actions).")
        bad_len_ap  = lambda ap: CobaException("An explicit action_prob was passed but it is not "
        "a two piece tuple. A valid format has the form (<action>,<prob>). We were given {ap}.")

        no_len = lambda item: not hasattr(item,'__len__')

        #possible std_pred:
            #pmf, action, [action,prob], {'pmf':...}, {'action':...}, {'action_prob':...}

        if isinstance(std_pred, dict):
            if 'pmf' in std_pred:
                pmf = std_pred['pmf']
                if not actions: raise no_act_pmf
                if no_len(pmf) or len(pmf) != len(actions): raise bad_len_pmf
                return 'PM*'
            if 'action' in std_pred:
                return 'AX*'
            if 'action_prob' in std_pred:
                ap = std_pred['action_prob']
                if no_len(ap) or len(ap) != 2:
                    raise bad_len_ap(ap)
                return 'AP*'

        if no_len(std_pred) or isinstance(std_pred,str):
            #action
            std_pred = [std_pred]
        elif len(std_pred) > 2:
            #pmf or action
            std_pred = [std_pred]
        elif len(std_pred) == 2:
            #pmf, action or [action,prob]
            if not actions:
                raise no_act_two
            elif any(std_pred[0] is a for a in actions):
                #[action,prob] (this should always be correct due to _safe_actions)
                #when could pred[0] be identified as action but it isn't?
                #see, https://stackoverflow.com/a/306353/1066291
                pass
            else:
                #pmf or action
                std_pred = [std_pred]

        #at this point pred should be
            #[pmf], [action], [action,prob]

        if len(std_pred) == 2:
            #only [action,prob] has two items
            return 'AP'

        #now it is [pmf] or [action]

        if actions == [] or actions is None:
            #a continuous action problem
            #so they had to return action
            return 'AX'

        if any(std_pred[0] is action for action in actions):
            return 'AX'

        if SafeLearner.possible_pmf(std_pred[0],actions):
            return 'PM'

        if SafeLearner.possible_action(std_pred[0],actions):
            return 'AX'

        raise unclear_format

    @staticmethod
    def possible_pmf(item, actions):
        try:
            return len(item)==len(actions) and isclose(sum(item), 1, abs_tol=.001) and all(i >=0 for i in item)
        except:
            return False

    @staticmethod
    def possible_action(item, actions):
        try:
            return item in actions or len(actions) == 0
        except:
            return False

    @staticmethod
    def batch_size(args):
        for arg in args:
            if is_batch(arg):
                return len(arg)

    @staticmethod
    def raise_if_not_valid_out(out,expected_len):
        def len_or_0(obj):
            return try_else(lambda: len(obj), 0)
        if out is None:
            raise CobaException("The given prediction was none and did not match the batch_size.")
        if all(isinstance(p,dict) for p in out) and out[0].keys() != out[-1].keys(): #pragma: no cover
            out = out[0]
        if isinstance(out,dict):
            is_valid = expected_len == len_or_0(next(iter(out.values())))
        else:
            is_valid = expected_len in [len_or_0(out),len_or_0(out[0])]
        if not is_valid:
            raise CobaException("The given prediction did not appear to match the batch_size.")

    def __init__(self, learner: Learner, seed:int=1) -> None:
        """Instantiate a SafeLearner.

        Args:
            learner: The learner we wish to make sure has the expected interface
        """

        self.learner = learner if not isinstance(learner, SafeLearner) else learner.learner
        self._rng    = CobaRandom(seed)
        self._method = {}

        self._pred_kwargs = None
        self._pred_batch  = None
        self._pred_format = None

        self._prev_actions = None
        self._safe_actions = None

    @property
    def full_name(self) -> str:
        """A user-friendly name created from a learner's params for reporting purposes."""

        params = dict(self.params)
        family = params.pop("family")

        if len(params) > 0:
            return f"{family}({','.join(f'{k}={v}' for k,v in params.items())})"
        else:
            return family

    @property
    def params(self) -> Mapping[str, Any]:
        try:
            params = self.learner.params
            params = params if not callable(params) else params()
            params = params if isinstance(params,dict) else {'params':str(params)}
        except AttributeError:
            params = {}

        if "family" not in params:
            params["family"] = self.learner.__class__.__name__

        return params

    @property
    def has_score(self) -> bool:
        try:
            self.learner.score(None,None,None)
        except Exception as ex:
            return "score" not in str(ex)
        else:
            return True

    def _method1(self,method,args,kwargs):
        return method(*args,**kwargs)

    def _method2(self,method,args,kwargs):
        pred = [ method(*a,**{k:v[i] for k,v in kwargs.items()}) for i,a in enumerate(zip(*args)) ]
        if not pred:
            raise CobaException(
                f"Something went wrong. No prediction was returned when using batch fallback methods "
                f"on the args: {args}. See the exception above for the reason why fallback methods "
                "were attempted."
            )
        return pred

    def _safe_call(self, key, method, args, kwargs = {}, has_out = True):
        if self._method.get(key) == 1:
            return self._method1(method, args, kwargs)
        elif self._method.get(key) == 2:
            return self._method2(method, args, kwargs)
        elif not any(map(is_batch,args)):
            self._method[key] = 1
            return self._method1(method,args,kwargs)
        else:
            expected_size = SafeLearner.batch_size(args)
            try:
                out = self._method1(method,args,kwargs)
                if has_out: SafeLearner.raise_if_not_valid_out(out,expected_size)
                self._method[key] = 1
                return out
            except Exception as outer_e:
                try:
                    out = self._method2(method,args,kwargs)
                    if has_out: SafeLearner.raise_if_not_valid_out(out,expected_size)
                    self._method[key] = 2
                    return out
                except Exception as inner_e:
                    raise inner_e from outer_e

    def _parse_pred(self, context, actions, pred):
        if self._pred_batch is None: # first call only
            predictor         = lambda X,A: self._safe_call('predict', self.learner.predict, (X,A))
            self._pred_batch  = SafeLearner.batch_order(predictor,pred,context,actions,self._method['predict'])
            self._pred_kwargs = SafeLearner.has_kwargs(pred,self._pred_batch)
            first_pred_row    = SafeLearner.first_row(pred,self._pred_batch,self._pred_kwargs)
            first_actions     = actions[0] if self._pred_batch != 'not' else actions
            self._pred_format = SafeLearner.pred_format(first_pred_row, first_actions, pred)

            if self._pred_format[:2] == "PM":
                CobaContext.logger.log(f"<<< WARNING >>> Support for PMF predictions from learners will be removed in the future.")
                CobaContext.logger.log(f"<<< WARNING >>> Please use coba.random.choicew(actions,pmf) to return an action instead.")

        if self._pred_batch == 'not':
            kwargs = pred[-1] if self._pred_kwargs else {}
            pred   = pred[ 0] if self._pred_kwargs and len(pred)==2 else pred

            if self._pred_format.endswith('*'):
                pred = list(pred.values())[0]

            if self._pred_format[:2] == 'PM':
                a,p = self._rng.choicew(actions,pred)

            if self._pred_format[:2] == 'AP':
                a,p = pred[:2]

            if self._pred_format[:2] == 'AX':
                a,p = pred,None

            return a,p,kwargs

        if self._pred_batch == 'row':
            kwargs = [ p[-1] if self._pred_kwargs else {} for p in pred ]
            kwargs = {k: [kw[k] for kw in kwargs] for k in kwargs[0] }
            pred   = [p[0] if len(p)==2 else p[:-1] for p in pred] if self._pred_kwargs else pred

            if self._pred_format.endswith('*'):
                pred = [list(p.values())[0] for p in pred]

            A,P = [],[]
            if self._pred_format[:2] == 'PM':
                A, P = list(map(list, zip(*map(self._rng.choicew,actions, pred))))

            if self._pred_format[:2] == 'AX':
                A = pred
                P = [None]*len(pred)

            if self._pred_format[:2] == 'AP':
                A,P = zip(*pred)

            return A,P,kwargs

        if self._pred_batch == 'col':
            kwargs = pred[-1] if self._pred_kwargs else {}
            pred   = pred[:-1] if self._pred_kwargs else pred

            if self._pred_format.endswith('*'):
                pred = list(pred.values())[0]

            if self._pred_format[:2] == 'PM':
                A, P = list(map(list, zip(*map(self._rng.choicew,actions, pred))))

            if self._pred_format[:2] == 'AX':
                A = pred
                P = [None]*len(pred)

            if self._pred_format[:2] == 'AP':
                if self._pred_format.endswith('*'):
                    A,P = zip(*pred)
                else:
                    A,P = pred

            return A,P,kwargs

    def score(self, context: Context, actions: Actions, action: Action) -> Prob:
        try:
            return self._safe_call('score', self.learner.score,(context,actions,action))
        except AttributeError as ex:
            if "'score'" in str(ex):
                raise CobaException(("The `score` method is not implemented for this learner."))
            raise

    def predict(self, context: Context, actions: Actions) -> Tuple[Action,Prob,Kwargs]:
        #this logic should guarantee that we can differentiate prediction formats
        #it allows us to "is" checks to see if a returned value "is" one of the actions
        if self._prev_actions != actions:
            self._prev_actions = actions
            all_safe = 0 not in actions and 1 not in actions
            make_safe = lambda a: float(a) if a in [0,1] else a
            self._safe_actions = actions if all_safe else [ make_safe(a) for a in actions]

        pred = self._safe_call('predict', self.learner.predict, (context,self._safe_actions))
        return self._parse_pred(context, self._safe_actions, pred)

    def learn(self, context, action, reward, probability, **kwargs) -> None:
        try:
            self._safe_call('learn', self.learner.learn, (context,action,reward,probability), kwargs, has_out=False)
        except TypeError as ex:
            if 'got an unexpected' in str(ex):
                raise CobaException("It appears that learner.predict returned kwargs but learner.learn did not accept them.") from ex
            if 'learn() missing' in str(ex):
                raise CobaException("It appears that learner.learn expected kwargs but learner.predict did not provide any.") from ex
            raise

    def __str__(self) -> str:
        return self.full_name

class SafeEvaluator(Evaluator):
    def __init__(self, evaluator: Union[Evaluator, Callable[[Environment,Learner], Union[Mapping, Iterable[Mapping]]]]) -> None:
        self.evaluator = evaluator if not isinstance(evaluator, SafeEvaluator) else evaluator.evaluator

    @property
    def params(self):
        try:
            params = self.evaluator.params
        except:
            params = {}

        if type(self.evaluator).__name__ == 'function':
            params['eval_type'] = self.evaluator.__name__
        else:
            params['eval_type'] = type(self.evaluator).__name__

        return params

    def evaluate(self, environment: Optional[Environment], learner: Optional[Learner]) -> Union[Mapping, Iterable[Mapping]]:
        if callable(self.evaluator):
            return self.evaluator(environment,learner)
        else:
            return self.evaluator.evaluate(environment,learner)



import sys
import json
import collections
import traceback

from pathlib import Path
from typing import Iterable, Dict, Any, Sequence, Union, Literal

from coba.exceptions import CobaException
from coba.registry import JsonMakerV1, CobaRegistry
from coba.utilities import coba_exit

from coba.context.cachers import Cacher
from coba.context.loggers import Logger

# To support class properties before python 3.9 we must implement our properties directly
# on a meta class. Using class properties rather than class variables is done to allow
# lazy loading. Lazy loading moves errors to execution time instead of import time where
# they are easier to debug as well as removing import time circular references.

# Creating a global configuration context to allow easy mocking and customization.
# In short, So long as the same modulename is always used to import and the import
# always occurs on the same thread I'm fairly confident this pattern will always work.
# In long, I'm somewhat unsure about this pattern for the following reasons:
#     > While there seems concensus that multi-import doesn't repeat (see [1]) it may if different modulenames are used (see [2])
#         [1] https://stackoverflow.com/a/19077396/1066291
#         [2] https://stackoverflow.com/q/13392038/1066291
#     > Python 3.7 added in explicit context management (see [3,4]) but this functionality is thread local only
#         [3] https://www.python.org/dev/peps/pep-0567/
#         [4] https://docs.python.org/3/library/contextvars.html

class ExperimentConfig:
    """Default configuration settings for Experiments.

    These can be overridden directly calling `config` on an Experiment.
    """

    def __init__(self, processes:int, maxchunksperchild:int, maxtasksperchunk:int, chunk_by: Literal["source","task"]):
        """Instantiate an ExperimentConfig."""
        self.processes        : int                      = processes
        self.maxchunksperchild: int                      = maxchunksperchild
        self.maxtasksperchunk : int                      = maxtasksperchunk
        self.chunk_by         : Literal["source","task"] = chunk_by

class CobaContext_meta(type):
    """Global execution context.

    CobaContext attributes can be set programatically or in a config file.
    """

    _api_keys      = None
    _cacher        = None
    _logger        = None
    _experiment    = None
    _search_paths  = [Path.home(), Path.cwd(), Path(sys.path[0])]
    _store         = {}
    _learning_info = {}

    def _load_file_configs(cls) -> Dict[str,Any]:
        config = {}

        for search_path in cls.search_paths:

            potential_coba_config = search_path / ".coba"

            if potential_coba_config.exists() and potential_coba_config.read_text().strip() != "":
                try:
                    file_config = json.loads(potential_coba_config.read_text())

                    if not isinstance(file_config, dict):
                        raise CobaException(f"Expecting a JSON object (i.e., {{}}).")

                    cls._resolve_and_expand_paths(file_config, str(search_path))

                    config.update(file_config)

                except Exception as e:
                    raise CobaException(f"{str(e).strip('.')} in {potential_coba_config}.")

        return config

    def _resolve_and_expand_paths(cls, config_dict: dict, current_dir:str):
        for key,item in config_dict.items():
            if isinstance(item, dict):
                cls._resolve_and_expand_paths(item, current_dir)

            if isinstance(item,str) and item.strip().startswith("~/"):
                config_dict[key] = str(Path(item).expanduser().resolve())

            if isinstance(item,str) and (item.strip().startswith("../") or item.strip().startswith("./")):
                config_dict[key] = str(Path(current_dir,item).resolve())

    _config_backing = None

    @property
    def _config(cls) -> Dict[str,Any]:

        if cls._config_backing is None:
            try:
                _raw_config: Dict[str,Any] = {
                    "api_keys"  : collections.defaultdict(lambda:None),
                    "cacher"    : { "DiskCacher": "~/.cache/coba"},
                    "logger"    : { "IndentLogger": "Console" },
                    "experiment": { "processes": 1, "maxchunksperchild": 0, "maxtasksperchunk":0, "chunk_by": "source" }
                }

                for key,value in cls._load_file_configs().items():
                    if key in ['api_keys', 'experiment']:

                        if key == "experiment" and "maxtasksperchild" in value:
                            value["maxchunksperchild"] = value["maxtasksperchild"]
                            del value["maxtasksperchild"]

                        _raw_config[key].update(value)
                    else:
                        _raw_config[key] = value

                cls._config_backing = {
                    'api_keys'  : _raw_config['api_keys'],
                    'cacher'    : JsonMakerV1(CobaRegistry.registry).make(_raw_config['cacher']),
                    'logger'    : JsonMakerV1(CobaRegistry.registry).make(_raw_config['logger']),
                    'experiment': ExperimentConfig(**_raw_config['experiment'])
                }
            except CobaException as e:
                messages = [
                    '',
                    "ERROR: An error occured while initializing CobaContext. Execution is unable to continue. Please see below for details:",
                    f"    > {e}",
                    ''
                ]
                coba_exit('\n'.join(messages))

            except Exception as e:
                messages = [
                    '',
                    "ERROR: An error occured while initializing CobaContext. Execution is unable to continue. Please see below for details:",
                    ''.join(traceback.format_tb(e.__traceback__)),
                    ''.join(traceback.TracebackException.from_exception(e).format_exception_only())
                ]
                coba_exit('\n'.join(messages))

        return cls._config_backing

    @property
    def api_keys(cls) -> Dict[str,str]:
        """Global API key collection."""
        cls._api_keys = cls._api_keys if cls._api_keys else cls._config['api_keys']
        return cls._api_keys

    @api_keys.setter
    def api_keys(cls, value: Dict[str,str]):
        cls._api_keys = value

    @property
    def cacher(cls) -> Cacher[str,Iterable[bytes]]:
        """Global caching strategy."""
        cls._cacher = cls._cacher if cls._cacher else cls._config['cacher']
        return cls._cacher

    @cacher.setter
    def cacher(cls, value: Union[Cacher,str]) -> None:
        cls._cacher = value

    @property
    def logger(cls) -> Logger:
        """Global logging strategy."""
        cls._logger = cls._logger if cls._logger else cls._config['logger']
        return cls._logger

    @logger.setter
    def logger(cls, value: Logger) -> None:
        cls._logger = value

    @property
    def store(cls) -> Dict[str,Any]:
        """Global store of objects.

        Remarks:
            This store will be marshalled to background
            processes when multiprocessing is used.
        """
        return cls._store

    @store.setter
    def store(cls, value:Dict[str,Any]) -> None:
        cls._store = value

    @property
    def experiment(cls) -> ExperimentConfig:
        """Global Experiment configurations."""
        cls._experiment = cls._experiment if cls._experiment else cls._config['experiment']
        return cls._experiment

    @property
    def search_paths(cls) -> Sequence[Path]:
        """Global search paths for config files."""
        return cls._search_paths

    @search_paths.setter
    def search_paths(cls, value:Sequence[Union[str,Path]]) -> None:
        cls._search_paths = [ Path(path) if isinstance(path,str) else path for path in value  ]

    @property
    def learning_info(cls) -> Dict[str,Any]:
        """Golbal Learner info.

        Remarks:
            When using coba evaluators values stored in learning info will be
            moved to evaluator results. This makes it possible to log information
            on learn/predict calls for later analysis on the performance of a learner.
        """
        cls._learning_info = cls._learning_info if cls._learning_info is not None else {}
        return cls._learning_info

class CobaContext(metaclass=CobaContext_meta):
    """To support class properties before python 3.9 we must implement our properties directly
       on a meta class. Using class properties rather than class variables is done to allow
       lazy loading. Lazy loading moves errors to execution time instead of import time making
       them easier to debug as well as removing the potential for import time circular references.
    """ #I'm not sure some of the claims in the docstring are still true...
    pass



from copy import deepcopy
from itertools import islice
from collections import defaultdict, Counter
from typing import Any, Iterable, Sequence, Optional, Tuple

from coba.context import CobaContext
from coba.utilities import peek_first
from coba.primitives import Source, Filter, Learner, Environment, Evaluator
from coba.safety import SafeLearner, SafeEnvironment, SafeEvaluator

from coba.results import Result

class Task:
    def __init__(self,
        env : Optional[Tuple[int,Environment]],
        lrn : Optional[Tuple[int,Learner,bool]],
        val : Optional[Tuple[int,Evaluator]],
        copy: bool = False) -> None:
        (self.env_id, self.env) = env or (None,None)
        (self.lrn_id, self.lrn) = lrn or (None,None)
        (self.val_id, self.val) = val or (None,None)
        self.copy   = copy

    def __eq__(self, o: object) -> bool:
        return isinstance(o,Task) \
        and self.env_id == o.env_id \
        and self.env    == o.env    \
        and self.lrn_id == o.lrn_id \
        and self.lrn    == o.lrn    \
        and self.val_id == o.val_id \
        and self.val    == o.val    \
        and self.copy   == o.copy   \

class MakeTasks(Source[Iterable[Task]]):

    def __init__(self,
        triples: Sequence[Tuple[Environment,Learner,Evaluator]],
        restored: Optional[Result] = None) -> None:

        self._triples = triples
        self._restored = restored or Result()

    def read(self) -> Iterable[Task]:

        #we rely on ids to make sure we don't do duplicate work. So long as self.evaluation_pairs
        #is always in the exact same order we should be fine. In the future we may want to consider.
        #adding a better check for environments other than assigning an index based on their order.

        envs = {}
        lrns = {}
        vals = {}

        restored_lrns = set(self._restored.learners['learner_id'])
        restored_envs = set(self._restored.environments['environment_id'])
        restored_vals = set(self._restored.evaluators['evaluator_id'])
        restored_outs = set(zip(*self._restored.interactions[['environment_id','learner_id','evaluator_id']]))

        learner_counts = Counter([l for _,l,_ in self._triples])

        for env, lrn, val in self._triples:

            if env not in envs:
                eid = len(envs)
                envs[env] = eid
                if eid not in restored_envs:
                    yield Task((eid,env),None,None)

            if lrn not in lrns:
                lid = len(lrns)
                lrns[lrn] = lid
                if lid not in restored_lrns:
                    yield Task(None,(lid,lrn),None)

            if val not in vals:
                vid = len(vals)
                vals[val] = vid
                if vid not in restored_vals:
                    yield Task(None,None,(vid,val))

            eid,lid,vid = (envs[env],lrns[lrn],vals[val])
            if val and (eid,lid,vid) not in restored_outs:
                yield Task((eid,env),(lid,lrn),(vid,val),copy=learner_counts[lrn]>1)

class ChunkTasks(Filter[Iterable[Task], Iterable[Sequence[Task]]]):

    def __init__(self, max_tasks: int = None) -> None:
        self._max_tasks = max_tasks or None

    def filter(self, items: Iterable[Task]) -> Iterable[Sequence[Task]]:
        return self._chunks(items)

    def _chunks(self, items: Iterable[Task]) -> Iterable[Sequence[Task]]:
        items  = list(items)
        chunks = defaultdict(list)

        tasks_sans_env = [t for t in items if not t.env ]
        tasks_with_env = [t for t in items if     t.env ]

        for task in tasks_with_env:
            chunks[self._get_last_chunk(task.env)].append(task)

        for task in tasks_sans_env:
            yield [task]

        for task in chunks.pop('not_chunked',[]):
            yield [task]

        chunks_sorter = lambda c: min([c.env_id for c in c])
        chunk_sorter  = lambda t: (t.env_id, t.lrn_id if t.lrn else -1)

        for chunk in sorted(chunks.values(), key=chunks_sorter):
            yield from self._max_chunker(sorted(chunk, key=chunk_sorter), self._max_tasks)

    def _get_last_chunk(self, env):
        from coba.environments import Chunk
        try:
            for pipe in reversed(list(env)):
                if isinstance(pipe, Chunk):
                    return pipe
        except:
            pass
        return 'not_chunked'

    def _max_chunker(self, chunk, max_tasks):
        chunk = iter(chunk)
        batch = list(islice(chunk,max_tasks))
        while batch != []:
            yield batch
            batch = list(islice(chunk,max_tasks))

class ProcessTasks(Filter[Iterable[Task], Iterable[Any]]):

    def filter(self, chunk: Iterable[Task]) -> Iterable[Any]:

        chunk = list(chunk)
        empty_envs = set()

        #We sort to make sure cached envs are grouped. This allows us to free envs from memory as we go.
        chunk = sorted(chunk, key=lambda item: self._env_ids(item)+self._lrn_ids(item), reverse=True)

        while chunk:
            try:
                task = chunk.pop()

                env_id,env = (task.env_id,task.env)
                lrn_id,lrn = (task.lrn_id,task.lrn)
                val_id,val = (task.val_id,task.val)

                is_e = env_id is not None
                is_l = lrn_id is not None
                is_v = val_id is not None

                if task.copy:
                    lrn = deepcopy(lrn)

                if is_e and not is_l and not is_v:
                    with CobaContext.logger.time(f"Peeking at Environment {env_id}..."):
                        #sometimes an environment can't know its parameters until we've
                        #read the first line. Therefore we call peek_first here and rely
                        #on side-effect behavior to update env.params
                        try:
                            peek_first(env.read())
                        except:
                            pass

                    with CobaContext.logger.time(f"Recording Environment {env_id} parameters..."):
                        yield ["T1", env_id, SafeEnvironment(env).params]

                if is_l and not is_e and not is_v:
                    with CobaContext.logger.time(f"Recording Learner {lrn_id} parameters..."):
                        yield ["T2", lrn_id, SafeLearner(lrn).params]

                if is_v and not is_e and not is_l:
                    with CobaContext.logger.time(f"Recording Evaluator {val_id} parameters..."):
                        yield ["T3", val_id, SafeEvaluator(val).params]

                if is_e and is_l and is_v and env_id not in empty_envs:
                    with CobaContext.logger.time(f"Evaluating Learner {lrn_id} on Environment {env_id}..."):
                        yield ["T4", (env_id, lrn_id, val_id), list(SafeEvaluator(val).evaluate(env,lrn))]
                        if hasattr(lrn,'finish') and task.copy: lrn.finish()

            except Exception as e:
                CobaContext.logger.log(e)

    def _env_ids(self, item: Task):
        return (item.env_id if item.env else -1,)

    def _lrn_ids(self, item: Task):
        return (item.lrn_id if item.lrn else -1,)



from typing import Mapping, Union, Iterable, Any

from coba.exceptions import CobaException
from coba.primitives import Filter, Source, Sink, Line

from coba.pipes.sources import SourceFilters
from coba.pipes.filters import FiltersFilter
from coba.pipes.sinks   import FiltersSink
from coba.pipes.lines   import SourceSink

class Foreach(Filter[Iterable[Any], Iterable[Any]], Sink[Iterable[Any]]):
    """A pipe that wraps an inner pipe and passes items to it one at a time."""

    def __init__(self, pipe: Union[Filter[Any,Any],Sink[Any]]):
        """Instantiate a Foreach pipe.

        Args:
            pipe: The pipe that we wish to pass a sequence of items one at a time.
        """
        self._pipe = pipe

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:
        """Filter the items using the inner pipe. This method only works if the inner pipe is a Filter."""
        yield from map(self._pipe.filter,items)

    def write(self, items: Iterable[Any]):
        """Write the items using the inner pipe. This method only works if the inner pipe is a Sink."""
        write = self._pipe.write
        for item in items:
            write(item)

    @property
    def params(self) -> Mapping[str,Any]:
        return self._pipe.params

    def __str__(self):
        return str(self._pipe)

class Pipes:
    @staticmethod
    def join(*pipes: Union[Source, Filter, Sink]) -> Union[Source, Filter, Sink, Line]:
        """Join a sequence of pipes into a single pipe.

        Args:
            pipes: a sequence of pipes.

        Returns:
            A single pipe that is a composition of the given pipes. The type of pipe returned
            is determined by the sequence given. A sequence of Filters will return a Filter. A
            sequence that begins with a Source and is followed by Filters will return a Source.
            A sequence that starts with Filters and ends with a Sink will return a Sink. A
            sequence that begins with a Source and ends with a Sink will return a completed pipe.
        """
        if len(pipes) == 0:
            raise CobaException("No pipes were passed to join.")

        if len(pipes) == 1 and hasattr(pipes[0],'read'):
            return SourceFilters(*pipes)

        if len(pipes) == 1 and hasattr(pipes[0],'filter'):
            return FiltersFilter(*pipes)

        if len(pipes) == 1 and hasattr(pipes[0],'write'):
            return FiltersSink(*pipes)

        first = pipes[0 ] if not isinstance(pipes[0 ], Foreach) else pipes[0 ]._pipe
        last  = pipes[-1] if not isinstance(pipes[-1], Foreach) else pipes[-1]._pipe

        if hasattr(first,'read') and hasattr(last,'write'):
            return SourceSink(*pipes)

        if hasattr(first,'read') and hasattr(last,'filter'):
            return SourceFilters(*pipes)

        if hasattr(first,'filter') and hasattr(last,'filter'):
            return FiltersFilter(*pipes)

        if hasattr(first,'filter') and hasattr(last,'write'):
            return FiltersSink(*pipes)

        raise CobaException("An unknown pipe was passed to join.")

def join(*pipes: Union[Source, Filter, Sink]) -> Union[Source, Filter, Sink, Line]:
    return Pipes.join(*pipes)



"""Learner implementations."""

from coba.learners.bandit     import BanditEpsilonLearner, BanditUCBLearner, FixedLearner, RandomLearner
from coba.learners.corral     import CorralLearner
from coba.learners.vowpal     import VowpalMediator
from coba.learners.vowpal     import VowpalLearner, VowpalEpsilonLearner, VowpalSoftmaxLearner, VowpalBagLearner, VowpalRndLearner
from coba.learners.vowpal     import VowpalCoverLearner, VowpalRegcbLearner, VowpalSquarecbLearner, VowpalOffPolicyLearner
from coba.learners.linucb     import LinUCBLearner
from coba.learners.lints      import LinTSLearner
from coba.learners.misguided  import MisguidedLearner
